/* Program for finding quadrature rules for triangles/tetrahedra
 *
 * Author: Lin-bo Zhang (zlb@lsec.cc.ac.cn)
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences.
 *
 * $Id: quadrule.c,v 1.144 2008/01/30 06:45:54 zlb Exp $ */

#define _GNU_SOURCE	/* for M_PIl */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <assert.h>
#include <time.h>
#if HAVE_FEENABLEEXCEPT
# include <fenv.h>
#endif  /* HAVE_FEENABLEEXCEPT */

#ifndef USE_MPI
# define	USE_MPI	0
#endif
#define USE_LONG_DOUBLE	1

#if USE_LONG_DOUBLE
typedef long double FLOAT;
# define Sqrt(a)	sqrtl((FLOAT)(a))
# define Pow(a)		powl((FLOAT)(a))
# define Atanh(a)	atanhl((FLOAT)(a))
# define Tanh(a)	tanhl((FLOAT)(a))
# define Fabs(a)	fabsl((FLOAT)(a))
# ifdef M_PIl
# define PI		M_PIl
# else
# define PI		3.1415926535897932384626433832795029L
# endif
#else	/* USE_LONG_DOUBLE */
typedef double FLOAT;
# define Sqrt(a)	sqrt((FLOAT)(a))
# define Pow(a)		pow((FLOAT)(a))
# define Atanh(a)	atanh((FLOAT)(a))
# define Tanh(a)	tanh((FLOAT)(a))
# define Fabs(a)	fabs((FLOAT)(a))
# define PI		M_PI
#endif	/* USE_LONG_DOUBLE */

#if USE_MPI
# include <mpi.h>
#endif	/* USE_MPI */

/* global parameters */

#define REPORT_TOL	1e-9	/* tolerance for reporting a solution */
#define ACCEPT_TOL	1e-12	/* tolerance for accepting a solution */
#define INDEPENDENT_W	0	/* != 0 ==> treat weights as independent vars */
#define LMDIF_MAXFEV	5000	/* MAXFEV=LMDIF_MAXFEV * (n + 1) in lmdif1() */
#define EPS		1e-18	/* convergence criterion for the NLS solver */

#define LEAST_SQUARE	1	/* linear least square solution for weights */
#if USE_LONG_DOUBLE
# define USE_LAPACK	0	/* don't use LAPACK for linear LS (precision) */
#else	/* USE_LONG_DOUBLE */
# define USE_LAPACK	1	/* use LAPACK for linear LS */
#endif	/* USE_LONG_DOUBLE */

#define ORTHOGONAL_BASIS 0	/* whether use orthogonal basis */
#define USE_CARDINAL	0	/* use cardinal functions (DON'T CHANGE!) */

#define ASYNCHRONOUS	1	/* asynchronous mode (use file to communi.) */

/* WARNING: TENSOLVE interface is not working yet! */
#define USE_TENSOLVE	0	/* != 0 ==> use TENSOLVE instead of MINPACK */

/* WARNING: genetic code not finished and currently broken */
#define GENETIC		0	/* != 0 ==> use G.A. to find initial approx. */

/* WARNING: code with SYMMETRY == 0 is broken */
#define SYMMETRY	1	/* != 0 ==> enforce symmetry */
#define SYMMETRY_BASIS	1	/* != 0 ==> enforce symmetry of basis funcs */

#define Random(a)	(rand() * (FLOAT)(a) / RAND_MAX)

typedef struct {
    FLOAT coeff0;	/* coefficient for a single term */
    FLOAT *coeff;	/* coefficients */
    int *o;		/* list of orders */
    int order;		/* order of the polynomial */
} POLY;

static int guess_max = 200;
static int nprocs = 1, myrank = 0;

static int plot = 0;
static int print = 0;

static int specified_stars = 0;
static int s3, s21, s111, s4, s31, s22, s211, s1111;
static int degree, dim, npolys, npolys0, neqns, nabsc, nwgts;
static int npoints = 0, nsols = 0;
static int *porders = NULL, *pdims = NULL;

static POLY **pbasis = NULL;		/* single term basis polynomials */
static int dummy[] = {0,0,0};
static POLY one_ = {1.0, NULL, dummy, 0};
#define ONE	(&one_)

#if USE_CARDINAL
static POLY **obasis = NULL;	/* orthogonal basis polynomials */
static int *oorders = NULL;
#endif	/* USE_CARDINAL */

#if ASYNCHRONOUS
# include <setjmp.h>
# include <unistd.h>
typedef enum {REMOVE_SYNC_FILE, CREATE_SYNC_FILE, CHECK_SYNC_FILE} sync_mode_t;
static jmp_buf jmp;

#if USE_MPI
static char *sync_fn = NULL;

static void
clean_sync_file(void)
{
    if (sync_fn == NULL)
	return;
    if (myrank == 0)
	unlink(sync_fn);
    free(sync_fn);
    sync_fn = NULL;
}
#endif	/* USE_MPI */

static int
sync_file(sync_mode_t mode)
{
#if USE_MPI
    static char *sync_fn = NULL;
    FILE *fp;

    if (nprocs <= 1)
	return 0;

    if (sync_fn == NULL) {
	sync_fn = malloc(PATH_MAX + 1);
	if (myrank == 0) {
	    char pn[MPI_MAX_PROCESSOR_NAME + 1], *p;
	    int len;
	    MPI_Get_processor_name(pn, &len);
	    pn[len] = '\0';
	    if ((p = strchr(pn, '.')) != NULL)
		*p = '\0';
	    len = strlen(pn);
	    sprintf(sync_fn, "quad_%s_%d.sync", len <= 16 ? pn : pn + len - 16,
			getpid());
	    unlink(sync_fn);
	}
	MPI_Bcast(sync_fn, PATH_MAX + 1, MPI_BYTE, 0, MPI_COMM_WORLD);
	atexit(clean_sync_file);
    }

    switch (mode) {
	case REMOVE_SYNC_FILE:
	    unlink(sync_fn);
	    return 0;
	case CREATE_SYNC_FILE:
	    fp = fopen(sync_fn, "w+t");
	    fprintf(fp, "done.\n");
	    fclose(fp);
	    return 0;
	case CHECK_SYNC_FILE:
	    if ((fp = fopen(sync_fn, "r")) != NULL)
		fclose(fp);
	    return (fp != NULL);
	default:
	    fprintf(stderr, "sync_file: invalid mode %d!\n", mode);
	    exit(1);
    }
#endif	/* USE_MPI */

    return 0;
}
#endif	/* ASYNCHRONOUS */

static inline FLOAT
Pow0(FLOAT x, int p)
{
    FLOAT d;

    assert(p >= 0);

    if (p == 0)
	return 1.;

    if (x == 0. || x == 1.)
	return x;

    if (x == -1.)
	return (p & 1) ? -1 : 1;

    while (!(p & 1)) {
	p >>= 1;
	x *= x;
    }
    d = x;
    while ((p >>= 1)) {
	x *= x;
	if (p & 1)
	    d *= x;
    }

    return d;
}

static inline FLOAT
Pow2(FLOAT x, int p, FLOAT y, int q)
{
    if (p == q) 
	return Pow0(x * y, p);
    else if (p > q)
	return (q == 0 ? 1. : Pow0(x * y, q)) * Pow0(x, p - q);
    else
	return (p == 0 ? 1. : Pow0(x * y, p)) * Pow0(y, q - p);
}

static inline FLOAT
Pow3(FLOAT x, int p, FLOAT y, int q, FLOAT z, int r)
{
    if (r <= p && r <= q)
	return (r == 0 ? 1. : Pow0(x * y * z, r)) * Pow2(x, p - r, y, q - r);
    else if (q <= p && q <= r)
	return (q == 0 ? 1. : Pow0(x * y * z, q)) * Pow2(x, p - q, z, r - q);
    else
	return (p == 0 ? 1. : Pow0(x * y * z, p)) * Pow2(y, q - p, z, r - p);
}

static FLOAT
integral(int p, int q, int r)
/* returns the integral $\int_S x^p y^q z^r$, $S$ is the standard simplex.
 *
 * The geeneral formula for integrals over a $(d+1)$-simplex is:
 * $$
 * 	\int_S \lambda^\alpha = \alpha! / (d + |\alpha|)!
 * $$
 * where $\lambda = (\lambda_0, \lambda_1, \ldots, \lambda_d)$,
 *	 $\alpha = (\alpha_0, \alpha_1, \ldots, \alpha_d)$,
 *	 $\alpha! = \alpha_0! \alpha_1! \cdots \alpha_d!$,
 * and:  $|\aplha| = \alpha_0 + \alpha_1 + \cdots + \alpha_d$ */
{
    FLOAT a;

#define EXCHG(p, q)	{int i = p; p = q; q = i;}
    if (dim == 2) {
	/* \int x^p y^q = p! q! / (2 + p + q)! */
	if (p < q) EXCHG(p, q);
	a = 1. / ((p + q + 1.) * (p + q + 2.)) ;
	for (; q > 0; q--)
	    a *= q / (FLOAT)(p + q);
    }
    else {
	/* \int x^p y^q z^r = p! q! r! / (3 + p + q + r)! */
	if (p < q) EXCHG(p, q);
	if (p < r) EXCHG(p, r);
	if (q < r) EXCHG(q, r);
	a = 1. / ((p + q + r + 1.) * (p + q + r + 2.) * (p + q + r + 3.)) ;
	for (; q > 0; q--)
	    a *= q / (FLOAT)(p + r + q);
	for (; r > 0; r--)
	    a *= r / (FLOAT)(p + r);
    }

    return a;
}

static void
poly_dump(POLY *p, FILE *f)
{
    (void)poly_dump;

    if (p->coeff != NULL) {
	int i, j = -1, *o;
	o = p->o;
	for (i = 0; i <= p->order; i++, o += dim) {
	    if (Fabs(p->coeff[i]) < EPS)
		continue;
	    fprintf(f, "%s %lg", j == -1 ? "" : " +", (double)p->coeff[i]);
	    j = 0;
	    if ((dim == 2 && o[0] + o[1] == 0) ||
		(dim == 3 && o[0] + o[1] + o[2] == 0))
		continue;
	    if (o[0] > 0)
		(o[0] == 1 ? fprintf(f, " x") : fprintf(f, " x^%d", o[0]));
	    if (o[1] > 0)
		(o[1] == 1 ? fprintf(f, " y") : fprintf(f, " y^%d", o[1]));
	    if (dim == 3 && o[2] > 0)
		(o[2] == 1 ? fprintf(f, " z") : fprintf(f, " z^%d", o[2]));
	}
	fprintf(f, "\n");
    }
    else {
	int *o = p->o + dim * p->order;
	fprintf(f, "%lg",  (double)p->coeff0);
	if (o[0] > 0)
	    (o[0] == 1 ? fprintf(f, " x") : fprintf(f, " x^%d", o[0]));
	if (o[1] > 0)
	    (o[1] == 1 ? fprintf(f, " y") : fprintf(f, " y^%d", o[1]));
	if (dim == 3 && o[2] > 0)
	    (o[2] == 1 ? fprintf(f, " z") : fprintf(f, " z^%d", o[2]));
	fprintf(f, "\n");
    }
}

static FLOAT
poly_eval(POLY *p, FLOAT x, FLOAT y, FLOAT z)
{
    FLOAT va;
    int i;

    if (p->coeff != NULL) {
	/* TODO: number of operations could be largely reduced */
	va = 0.;
	if (dim == 2) {
	    for (i = 0; i <= p->order; i++) {
		if (Fabs(p->coeff[i]) < EPS)
		    continue;
		va += p->coeff[i] * Pow2(x, p->o[i + i],
					 y, p->o[i + i + 1]);
	    }
	}
	else {
	    for (i = 0; i <= p->order; i++) {
		if (Fabs(p->coeff[i]) < EPS)
		    continue;
		va += p->coeff[i] * Pow3(x, p->o[3 * i],
					 y, p->o[3 * i + 1],
					 z, p->o[3 * i + 2]);
	    }
	}
    }
    else {
	if (dim == 2) {
	    va = p->coeff0 * Pow2(x, p->o[2 * p->order],
				  y, p->o[2 * p->order + 1]);
	}
	else {
	    va = p->coeff0 * Pow3(x, p->o[3 * p->order],
				  y, p->o[3 * p->order + 1],
				  z, p->o[3 * p->order + 2]);
	}
    }

    return va;
}

static FLOAT
poly_eval_d(POLY *p, int which, FLOAT x, FLOAT y, FLOAT z)
/* evaluates derivative of 'p' with respect to 'which' (of x, y or z) */
{
    FLOAT va;
    int i;
    int a = 0, b = 0, c = 0;

    switch (which) {
	case 0: a = 1; break;
	case 1: b = 1; break;
	case 2: c = 1; break;
    }

    if (p->coeff != NULL) {
	/* TODO: number of operations could be largely reduced */
	va = 0.;
	if (dim == 2) {
	    for (i = 0; i <= p->order; i++) {
		if (Fabs(p->coeff[i]) < EPS || p->o[i + i + which] == 0)
		    continue;
		va += p->o[i + i + which] * p->coeff[i] *
			Pow2(x, p->o[i + i] - a, y, p->o[i + i + 1] - b);
	    }
	}
	else {
	    for (i = 0; i <= p->order; i++) {
		if (Fabs(p->coeff[i]) < EPS || p->o[3 * i + which] == 0)
		    continue;
		va += p->o[3 * i + which] * p->coeff[i] *
			Pow3(x, p->o[3 * i] - a, y, p->o[3 * i + 1] - b,
			     z, p->o[3 * i + 2] - c);
	    }
	}
	return va;
    }
    else {
	if (dim == 2) {
	    if (Fabs(p->coeff0) < EPS || p->o[2 * p->order + which] == 0)
		return 0.;
	    return p->o[2 * p->order + which] * p->coeff0 *
			Pow2(x, p->o[2 * p->order + 0] - a,
			     y, p->o[2 * p->order + 1] - b);
	}
	else {
	    if (Fabs(p->coeff0) < EPS || p->o[3 * p->order + which] == 0)
		return 0.;
	    return p->o[3 * p->order + which] * p->coeff0 *
			Pow3(x, p->o[3 * p->order + 0] - a,
			     y, p->o[3 * p->order + 1] - b,
			     z, p->o[3 * p->order + 2] - c);
	}
    }
}

static void
poly_scalar(POLY *p, FLOAT s)
{
    int i;

    if (p->coeff != NULL) {
	for (i = 0; i <= p->order; i++)
	    p->coeff[i] *= s;
    }
    else {
	p->coeff0 *= s;
    }
}

static void
poly_axpy(POLY *p, POLY *q, FLOAT s)
{
    int i;

    (void)poly_axpy;	/* to avoid 'unused ...' warning */

    assert(p->order >= q->order && p->o == q->o);

    if (p->coeff == NULL) {
	if ((i = p->order) == q->order) {
	    p->coeff0 += s * (q->coeff == NULL ? q->coeff0 : q->coeff[i]);
	    return;
	}
	p->coeff = calloc(i + 1, sizeof(*p->coeff));
	p->coeff[i] = p->coeff0;
    }

    if (q->coeff != NULL) {
	for (i = 0; i <= q->order; i++)
	    p->coeff[i] += s * q->coeff[i];
    }
    else {
	p->coeff[q->order] += s * q->coeff0;
    }
}

static POLY *
poly_new(int order, FLOAT a, int *orders)
{
    POLY *p = malloc(sizeof(POLY));

    p->order = order;
    p->coeff = NULL;
    p->o = orders;
    p->coeff0 = a;

    return p;
}

static POLY *
poly_dup(POLY *poly)
{
    POLY *p = malloc(sizeof(POLY));

    (void)poly_dup;

    memcpy(p, poly, sizeof(*p));
    if (p->coeff != NULL) {
	free(p->coeff);
	p->coeff = malloc((p->order + 1) * sizeof(*p->coeff));
	memcpy(p->coeff, poly->coeff, (p->order + 1) * sizeof(*p->coeff));
    }

    return p;
}

static void
poly_free(POLY *p)
{
    if (p->coeff != NULL)
	free(p->coeff);
    free(p);
}

static FLOAT
poly_dot(POLY *p, POLY *q)
{
    FLOAT res;
    int i, j, i0, j0, n, m, l = 0;
    FLOAT d, a, b;

    assert(q->order == 0 || q->o == p->o);

    i0 = (p->coeff == NULL ? p->order : 0);
    j0 = (q->coeff == NULL ? q->order : 0);

    res = 0.;
    for (i = i0; i <= p->order; i++) {
	if (Fabs(a = (p->coeff == NULL ? p->coeff0 : p->coeff[i])) < EPS)
	    continue;
	n = p->o[dim * i];
	m = p->o[dim * i + 1];
	if (dim == 3)
	   l = p->o[dim * i + 2];
	d = 0.;
	for (j = j0; j <= q->order; j++) {
	    if (Fabs(b = (q->coeff == NULL ? q->coeff0 : q->coeff[j])) < EPS)
		continue;
	    d += integral(n + p->o[dim * j], m + p->o[dim * j + 1],
			  dim == 2 ? 0 : l + p->o[dim * j + 2]) * b;
	}
	res += d * a;
    }

    return res;
}

static int
abscissas_2D(const FLOAT D[], FLOAT x[][dim], int wmap[])
/* computes the x[][2] array (and wmap[]) for a given set of free abscissas */
{
#if SYMMETRY
    FLOAT a, c, b;
    int i, j, k, l;

    j = k = l = 0;
    if (s3 == 1) {
	x[j][0] = x[j][1] = 1. / (dim + 1.L);
	wmap[j++] = l++;
    }
    for (i = 0; i < s21; i++) {
	a = D[k++];
	b = 1. - a - a;
	x[j][0] = a; x[j][1] = a;
	wmap[j++] = l;
	x[j][0] = a; x[j][1] = b;
	wmap[j++] = l;
	x[j][0] = b; x[j][1] = a;
	wmap[j++] = l++;
    }
    for (i = 0; i < s111; i++) {
	a = D[k++];
	b = D[k++];
	c = 1. - a - b;
	x[j][0] = a; x[j][1] = b;
	wmap[j++] = l;
	x[j][0] = b; x[j][1] = a;
	wmap[j++] = l;
	x[j][0] = a; x[j][1] = c;
	wmap[j++] = l;
	x[j][0] = c; x[j][1] = a;
	wmap[j++] = l;
	x[j][0] = b; x[j][1] = c;
	wmap[j++] = l;
	x[j][0] = c; x[j][1] = b;
	wmap[j++] = l++;
    }
#else	/* SYMMETRY */
    int k, j;

    k = 0;
    for (j = 0; j < nwgts; j++) {
	x[j][0] = D[k++];
	x[j][1] = D[k++];
	wmap[j] = j;
    }
#endif	/* SYMMETRY */

    return 0;
}

static int
abscissas_3D(const FLOAT D[], FLOAT x[][dim], int wmap[])
/* computes the abscissas x[][3] (and wmap[]) for given set of free abscissas */
{
#if SYMMETRY
    FLOAT a, b, c, d;
    int i, j, k, l;

#define Perm21(a,b)				\
	x[j][0] = a; x[j][1] = a; x[j][2] = b;	\
        wmap[j++] = l;				\
        x[j][0] = a; x[j][1] = b; x[j][2] = a;	\
        wmap[j++] = l;				\
        x[j][0] = b; x[j][1] = a; x[j][2] = a;	\
        wmap[j++] = l;

#define Perm111(a,b,c)				\
	x[j][0] = a; x[j][1] = b; x[j][2] = c;	\
	wmap[j++] = l;				\
	x[j][0] = a; x[j][1] = c; x[j][2] = b;	\
	wmap[j++] = l;				\
	x[j][0] = b; x[j][1] = c; x[j][2] = a;	\
	wmap[j++] = l;				\
	x[j][0] = b; x[j][1] = a; x[j][2] = c;	\
	wmap[j++] = l;				\
	x[j][0] = c; x[j][1] = a; x[j][2] = b;	\
	wmap[j++] = l;				\
	x[j][0] = c; x[j][1] = b; x[j][2] = a;	\
	wmap[j++] = l;

    j = k = l = 0;
    if (s4 == 1) {
	x[j][0] = x[j][1] = x[j][2] = 1. / (dim + 1.L);
	wmap[j++] = l++;
    }
    for (i = 0; i < s31; i++) {
	a = D[k++];
	b = 1. - a - a - a;
	x[j][0] = a; x[j][1] = a; x[j][2] = a;
	wmap[j++] = l;
	Perm21(a, b)
	l++;
    }
    for (i = 0; i < s22; i++) {
	a = D[k++];
	b = .5 - a;
	Perm21(a,b)
	Perm21(b,a)
	l++;
    }
    for (i = 0; i < s211; i++) {
	a = D[k++];
	b = D[k++];
	c = 1. - b - a - a;
	Perm21(a,b)
	Perm21(a,c)
	Perm111(a,b,c)
	l++;
    }
    for (i = 0; i < s1111; i++) {
	a = D[k++];
	b = D[k++];
	c = D[k++];
	d = 1. - a - b - c;
	Perm111(a,b,c)
	Perm111(a,b,d)
	Perm111(a,c,d)
	Perm111(b,c,d)
	l++;
    }
#undef Perm21
#undef Perm111
#else	/* SYMMETRY */
    int k, j;

    k = 0;
    for (j = 0; j < nwgts; j++) {
	x[j][0] = D[k++];
	x[j][1] = D[k++];
	x[j][2] = D[k++];
	wmap[j] = j;
    }
#endif	/* SYMMETRY */

    return 0;
}

static void
poly_gen(void)
/* generates an orthogonal basis for polynomials of degree 'degree' */
{
    POLY *p;
    int *o;
    int i, j, k, n;
    FLOAT a;

    pdims = malloc((degree + 1) * sizeof(*pdims));

#if 0
#if SYMMETRY
    pdims[0] = 1;
    for (i = 1; i <= degree; i++) {
	pdims[i] = pdims[i - 1] + (i / 2 + 1);
    }
#else	/* SYMMETRY */
    TODO: don t know how to compute pdims[] for dim == 3
#endif	/* SYMMETRY */
#endif

    if (dim == 2)
	npolys = (degree + 1) * (degree + 2) / 2;
    else
	npolys = (degree + 1) * (degree + 2) * (degree + 3) / 6;

    assert(npoints <= npolys);

    o = porders = malloc(dim * npolys * sizeof(*porders));

    if (dim == 2) {
	for (n = 0; n <= degree; n++) {
#if SYMMETRY && SYMMETRY_BASIS
	    for (i = n; i >= (n + 1) / 2; i--)
#else
	    for (i = n; i >= 0; i--)
#endif
	    {
		o[0] = i;
		o[1] = n - i;
		o += dim;
	    }
	    if (n >= degree) {
		pdims = realloc(pdims, (n + 1) * sizeof(*pdims));
	    }
	    pdims[n] = (o - porders) / dim;
	}
    }
    else {
	for (n = 0; n <= degree; n++) {
	    for (i = n; i >= 0; i--) {
		for (j = n - i; j >= 0; j--) {
#if SYMMETRY && SYMMETRY_BASIS
		    if (j > i || (k = n - i - j) > j)
			continue;
#else	/* SYMMETRY */
		    k = n - i - j;
#endif	/* SYMMETRY */
		    o[0] = i;
		    o[1] = j;
		    o[2] = k;
		    o += dim;
		}
	    }
	    pdims[n] = (o - porders) / dim;
	}
    }

#if SYMMETRY && SYMMETRY_BASIS
    npolys = pdims[degree];
    porders = realloc(porders, dim * npolys * sizeof(*porders));
#else
    assert((o - porders) / dim == npolys);
#endif

    /* Now generate the basis polynomials, and an orthonormal basis by
     * Gram-Schmidt procedure */

#if !ORTHOGONAL_BASIS
    pbasis = malloc(npolys * sizeof(*pbasis));
    p = pbasis[0] = poly_new(0, (FLOAT)1.0, porders);
    a = poly_dot(p, ONE);
    poly_scalar(p, 1./a);
    for (i = 1; i < npolys; i++) {
	p = pbasis[i] = poly_new(i, (FLOAT)1.0, porders);
	/* normalize the integral to 1 */
	a = poly_dot(p, ONE);
	poly_scalar(p, 1./a);
    }
#else	/* !ORTHOGONAL_BASIS */
    pbasis = malloc(npolys * sizeof(*pbasis));
    for (i = 0; i < npolys; i++) {
	p = pbasis[i] = poly_new(i, (FLOAT)1.0, porders);
    }
    /* orthogonize */
    for (j = 0; j < npolys; j++) {
	p = pbasis[j];
	/* normalize */
	a = poly_dot(p, p);
	poly_scalar(p, Sqrt(1./a));
	for (i = j + 1; i < npolys; i++) {
	    a = poly_dot(p, pbasis[i]);
	    poly_axpy(pbasis[i], p, -a);
	}
    }
    a = poly_dot(p = pbasis[0], ONE);
    poly_scalar(p, 1./a);
#endif	/* !ORTHOGONAL_BASIS */

#if USE_CARDINAL
#if ORTHOGONAL_BASIS
    oorders = porders;
    obasis = pbasis;
#endif	/* ORTHOGONAL_BASIS */
#define NCARD	(3 * npoints < npolys0 ? 3 * npoints : npolys0)
    k = NCARD;
    o = oorders = malloc(dim * k * sizeof(*oorders));
    if (dim == 2) {
	for (n = 0; ; n++) {
	    for (i = n; i >= 0; i--) {
		o[0] = i;
		o[1] = n - i;
		o += dim;
		if ((o - oorders) / dim >= k)
		    goto label;
	    }
	}
    }
    else {
	for (n = 0; ; n++) {
	    for (i = n; i >= 0; i--) {
		for (j = n - i; j >= 0; j--) {
		    o[0] = i;
		    o[1] = j;
		    o[2] = n - i - j;
		    o += dim;
		    if ((o - oorders) / dim >= k)
			goto label;
		}
	    }
	}
    }

label:
    k = NCARD;
    obasis = malloc(k * sizeof(*obasis));
    p = obasis[0] = poly_new(0, (FLOAT)1.0, oorders);
    a = poly_dot(p, p);
    poly_scalar(p, Sqrt(1./a));
    for (i = 1; i < k; i++) {
	p = obasis[i] = poly_new(i, (FLOAT)1.0, oorders);
	/* orthogonize */
	for (j = 0; j < i; j++) {
	    a = poly_dot(p, obasis[j]);
	    poly_axpy(p, obasis[j], -a);
	}
	/* normalize */
	a = poly_dot(p, p);
	poly_scalar(p, Sqrt(1./a));
    }

    /* scale the integral of first orgogonal polynomial to one */
    a = poly_dot(p = obasis[0], ONE);
    poly_scalar(p, 1./a);
#endif	/* USE_CARDINAL */

    return;
}

static int
linear_solver(int m, int n, int nb, int lda, FLOAT A[][lda])
{
    int i, j, k, p, q, nfree = 0;
    int *perm;
    FLOAT d, t, *tmp;
# define a(i,j) (A[i][j])

#if LEAST_SQUARE
    /* compute least square solution */
#if USE_LAPACK
    double *B, *WORK;
    int LWORK = m + (m + nb) * 1024, info;
    extern void dgels_(char *TRANS, int *M, int *N, int *NRHS,
			double *A, int *LDA, double *B, int *LDB,
			double *WORK, int *LWORK, int *INFO, int strlen);
    B = malloc(m * nb * sizeof(*B));
    WORK = malloc(LWORK * sizeof(*WORK));
    for (i = 0; i < m; i++) {
	for (j = 0; j < nb; j++)
	    B[i + j * m] = A[i][n + j];
    }
#if USE_LONG_DOUBLE
    {
	double (*A0)[lda] = calloc(m, sizeof(*A0));
	for (i = 0; i < m; i++)
	    for (j = 0; j < lda; j++)
		A0[i][j] = A[i][j];
	dgels_("T", &n, &m, &nb, (double *)A0, &lda, B, &m, WORK, &LWORK,
		&info, 1);
	for (i = 0; i < m; i++)
	    for (j = 0; j < lda; j++)
		A[i][j] = A0[i][j];
	free(A0);
    }
#else	/* USE_LONG_DOUBLE */
    dgels_("T", &n, &m, &nb, (double *)A, &lda, B, &m, WORK, &LWORK, &info, 1);
#endif	/* USE_LONG_DOUBLE */
    for (i = 0; i < n; i++) {
	for (j = 0; j < nb; j++)
	    A[i][n + j] = B[i + j * m];
    }
    free(B);
    free(WORK);
    return 0;
#else	/* USE_LAPACK */
    FLOAT *B;
    B = calloc(n * lda, sizeof(B[0]));
    for (k = 0; k < m; k++)
	for (i = 0; i < n; i++)
	    for (j = 0; j < n + nb; j++)
		B[i * lda + j] += A[k][i] * A[k][j];
    for (i = 0; i < n; i++)
	for (j = 0; j < n + nb; j++)
	    A[i][j] = B[i * lda + j];
    m = n;
    free(B);
#endif	/* USE_LAPACK */
#endif	/* LEAST_SQUARE */

    perm = malloc(n * sizeof(*perm));
    for (i = 0; i < n; i++)
	perm[i] = i;

    for (j = 0; j < n; j++) {
	/* find pivot */
	d = Fabs(a(j,j));
	p = q = j;
	for (i = j; i < m; i++) {
	    for (k = j; k < n; k++) {
		if (d < (t = Fabs(a(i,k)))) {
		    d = t;
		    p = i;
		    q = k;
		}
	    }
	}
#if USE_LONG_DOUBLE
	if (d < EPS * 1e-12)
#else	/* USE_LONG_DOUBLE */
	if (d < EPS * 1e-08)
#endif	/* USE_LONG_DOUBLE */
	{
	    nfree = n - j;
	    break;
	}
	if (p != j) {
	    /* swap rows j and k */
	    for (i = j; i < n + nb; i++) {
		t = a(j,i);
		a(j,i) = a(p,i);
		a(p,i) = t;
	    }
	}
	if (q != j) {
	    /* swap columns q and j */
	    for (k = 0; k < m; k++) {
		t = a(k, j);
		a(k, j) = a(k, q);
		a(k, q) = t;
	    }
	    i = perm[j];
	    perm[j] = perm[q];
	    perm[q] = i;
	}
	d = 1.0 / a(j,j);
	for (i = j + 1; i < n + nb; i++)
	    a(j,i) *= d;
	for (i = j + 1; i < m; i++) {
	    d = a(i,j);
	    for (k = j + 1; k < n + nb; k++)
		a(i,k) -= d * a(j,k);
	}
#if 1
	/* to facilitate debugging */
	a(j,j) = 1.;
	for (i = j + 1; i < m; i++)
	    a(i, j) = 0.;
#endif
    }

#if 0
	for (i = 0; i < m; i++) {
	    for (k = 0; k < n + 1; k++)
		fprintf(stderr, " %lg", (double)a(i,k));
	    fprintf(stderr, "\n");
	}
	/* check consistency of the equations */
	for (i = j; i < m; i++) {
	    if (Fabs(a(i, n)) >= EPS)
		return -1;
	}
#endif

    for (; j < n; j++)
	for (i = 0; i < nb; i++)
	    a(j, n + i) = 0.;

    for (j = n - 2; j >= 0; j--) {
	if (a(j,j) == 0.)
	    continue;
	for (i = j + 1; i < n; i++) {
	    for (k = n; k < n + nb; k++)
		a(j,k) -= a(j,i) * a(i,k);
	}
    }

    /* reorder the solution */
    tmp = malloc(n * sizeof(*tmp));
    for (i = n; i < n + nb; i++) {
	for (j = 0; j < n; j++)
	    tmp[perm[j]] = a(j, i);
	for (j = 0; j < n; j++)
	    a(j, i) = tmp[j];
    }
    free(tmp);

#undef a

    free(perm);
    return nfree;
}

#define USE_EXTREMAL_MEASURE	1
#define USE_INTERVALS		1
#define EXTREMAL_TRUNC		1e8

#if USE_INTERVALS

#define Perm3(a)        {a,a,a}
#define Dup3(w)         w
#define Perm21(a)       {a,a,1.-(a)-(a)}, {a,1.-(a)-(a),a}, {1.-(a)-(a),a,a}
#define Dup21(w)        w,w,w
#define Perm111(a,b)    {a,b,1.-(a)-(b)}, {a,1.-(a)-(b),b}, \
                        {b,a,1.-(a)-(b)}, {b,1.-(a)-(b),a}, \
                        {1.-(a)-(b),a,b}, {1.-(a)-(b),b,a}
#define Dup111(w)       w, w, w, w, w, w

/* 2D order 21 quadrature */
static FLOAT q_wts_2d[] = {
   Dup3(.02755999901737280567036428319984L),
   Dup21(.00327177164433417355412331988569L),
   Dup21(.00033522182193620438013546198443L),
   Dup21(.02205694663684318233535538889650L),
   Dup21(.02345762343122315902120690428678L),
   Dup21(.01176304901352800200621935870780L),
   Dup21(.01177857330753613281169605792613L),
   Dup111(.00227363040370311795915961787461L),
   Dup111(.00260387926598430901332115769841L),
   Dup111(.00463689207664425329134969311348L),
   Dup111(.00479688913117448706708110315036L),
   Dup111(.00571239045834759992270242810996L),
   Dup111(.00586084820867565010596626734524L),
   Dup111(.00940985776160413576012897789175L),
   Dup111(.01341231037151824632356079066506L),
   Dup111(.01571448883893753934172416324569L),
   Dup111(.01685980960794022658208936541953L),
   Dup111(.02138726471066724259277668911576L),
   Dup111(.02307347973420729737404411999326L)
};
static FLOAT q_pts_2d[][3] = {
   Perm3(.33333333333333333333333333333334L),
   Perm21(.04663974321501607479102430456662L),
   Perm21(.00355243919222250540042258171995L),
   Perm21(.20093771283185275362679788039499L),
   Perm21(.43765799038487432767692475040177L),
   Perm21(.38642155519549413716360405911677L),
   Perm21(.09549353103357583018866566257700L),
   Perm111(.95535482737294929469872570525395L,.03585527971776696834378129433497L),
   Perm111(.88652648790471610158911613485854L,.10823297450176790796300041060479L),
   Perm111(.78415203017701968847759819411523L,.20757204569459500783680568941600L),
   Perm111(.88270435625734606482401299518369L,.08581194897250000000000000000001L),
   Perm111(.66887782338253294738995138578902L,.32160710055498725434525230237642L),
   Perm111(.55201406712059379531184167949040L,.43799995431135144964688647799089L),
   Perm111(.79749310721474974652890063332828L,.16199749337338213888455380485838L),
   Perm111(.67747342805605078697487213263116L,.27454252387181407583630635649370L),
   Perm111(.54298496223437341547735952824134L,.40534724466673349360079873611735L),
   Perm111(.70541131168720444632770395072787L,.18777386155395484286218243953892L),
   Perm111(.57478172973482054342939017538145L,.30571229906428371583008680792599L),
   Perm111(.47178645433215042009513165241743L,.31213602566733864919506732062676L)
};

#undef Perm3
#undef Dup3
#undef Perm21
#undef Dup21
#undef Perm111
#undef Dup111

static int den_n = 100;
static FLOAT *den = NULL;

static void
density_2D(void)
/* computes:
 *	\int_{1/3-a}^{1/3+2a} \int_{1/3-a}^{2/3+a-x}
 *		1/sqrt{x y (1 - x - y)} dy dx */
{
    int i, j;
    FLOAT a, x0, y0, x, y, dx, dy, q;

    if (den != NULL)
	return;
    den = malloc((den_n + 1) * sizeof(*den));

    for (j = 0; j <= den_n; j++) {
	a = j / (3.L * den_n);
	if (j == 0) {
	    den[j] = 0.;
	    continue;
	}
	if (j == den_n) {
	    den[j] = 2. * PI;
	    continue;
	}
	x0 = y0 = 1./3.L - a;
	dx = dy = 3. * a;
	q = 0.;
	for (i = 0; i < sizeof(q_wts_2d) / sizeof(*q_wts_2d); i++) {
	    x = x0 + q_pts_2d[i][0] * dx;
	    y = y0 + q_pts_2d[i][1] * dy;
	    q += 1./Sqrt(x * y * (1. - x - y)) * q_wts_2d[i];
	}
	den[j] = q * 0.5 * dx * dy;
    }
}

#define Perm4(a)	{a,a,a,a}
#define Dup4(w)		w
#define Perm31(a)	{a,a,a,1.-3.*(a)}, {a,a,1.-3.*(a),a}, \
			{a,1.-3.*(a),a,a}, {1.-3.*(a),a,a,a}
#define Dup31(w)	w,w,w,w
#define Perm22(a)	{a,a,.5-(a),.5-(a)}, {a,.5-(a),a,.5-(a)}, \
			{a,.5-(a),.5-(a),a}, {.5-(a),a,.5-(a),a}, \
			{.5-(a),a,a,.5-(a)}, {.5-(a),.5-(a),a,a}
#define Dup22(w)	w,w,w,w,w,w
#define Perm211(a,b)	{a,a,b,1.-(a)-(a)-(b)}, {a,a,1.-(a)-(a)-(b),b}, \
			{a,b,a,1.-(a)-(a)-(b)}, {a,b,1.-(a)-(a)-(b),a}, \
			{a,1.-(a)-(a)-(b),a,b}, {a,1.-(a)-(a)-(b),b,a}, \
			{b,a,a,1.-(a)-(a)-(b)}, {b,a,1.-(a)-(a)-(b),a}, \
			{b,1.-(a)-(a)-(b),a,a}, {1.-(a)-(a)-(b),a,a,b}, \
			{1.-(a)-(a)-(b),a,b,a}, {1.-(a)-(a)-(b),b,a,a}
#define Dup211(w)	w,w,w,w,w,w,w,w,w,w,w,w
#define Perm0111(p,a,b,c)	{p,a,b,c}, {p,a,c,b}, {p,b,a,c}, \
				{p,b,c,a}, {p,c,a,b}, {p,c,b,a}
#define Perm1111(a,b,c) Perm0111(a,b,c,1.-(a)-(b)-(c)), \
			Perm0111(b,a,c,1.-(a)-(b)-(c)), \
			Perm0111(c,a,b,1.-(a)-(b)-(c)), \
			Perm0111(1.-(a)-(b)-(c),a,b,c)
#define Dup1111(w)	Dup111(w), Dup111(w), Dup111(w), Dup111(w)

#if 0
/* 3D order 10 quadrature */
static FLOAT q_wts_3d[] = {
    Dup4(0.0473997735560207L),
    Dup31(0.02693705999226870L),
    Dup31(0.009869159716793383L),
    Dup211(0.0003619443443392536L),
    Dup211(0.01013587167975579L),
    Dup211(0.01139388122019523L),
    Dup211(0.006576147277035904L),
    Dup211(0.02573973198045607L),
    Dup211(0.0129070357988620L)
};
static FLOAT q_pts_3d[][4] = {
    Perm4(0.25L),
    Perm31(0.3122500686951886L),
    Perm31(0.1143096538573462L),
    Perm211(0.00613800882479075L, 0.9429887673452049L),
    Perm211(0.03277946821644267L, 0.3401847940871076L),
    Perm211(0.4104307392189655L, 0.1654860256196111L),
    Perm211(0.03248528156482305L, 0.1338521522120095L),
    Perm211(0.1210501811455894L, 0.4771903799042804L),
    Perm211(0.1749793421839390L, 0.6280718454753660L)
};

#undef Perm4
#undef Dup4
#undef Perm31
#undef Dup31
#undef Perm22
#undef Dup22
#undef Perm211
#undef Dup211
#undef Perm0111
#undef Perm1111
#undef Dup1111

static void
density_3D(void)
/* computes:
 *	\int_{1/4-a}^{1/4+3a} \int_{1/4-a}^{1/2+2a-x} \int_{1/4-a}^{3/4+a-x-y}
 *		1/sqrt{x y z (1 - x - y - z)} dz dx dy */
{
    int i, j;
    FLOAT a, x0, y0, z0, x, y, z, dx, dy, dz, q;

    if (den != NULL)
	return;
    den = malloc((den_n + 1) * sizeof(*den));

    for (j = 0; j <= den_n; j++) {
	a = j / (4.L * den_n);
	if (j == 0) {
	    den[j] = 0.;
	    continue;
	}
	if (j == den_n) {
	    den[j] = PI * PI;
	    continue;
	}
	x0 = y0 = z0 = 1./4.L - a;
	dx = dy = dz = 4. * a;
	q = 0.;
	for (i = 0; i < sizeof(q_wts_3d) / sizeof(*q_wts_3d); i++) {
	    x = x0 + q_pts_3d[i][0] * dx;
	    y = y0 + q_pts_3d[i][1] * dy;
	    z = z0 + q_pts_3d[i][2] * dz;
	    q += 1./Sqrt(x * y * z * (1. - x - y - z)) * q_wts_3d[i];
	}
	den[j] = q * dx * dy * dz / 6.L;
    }
}
#endif	/* 0 */

#endif	/* USE_INTERVALS */

static void
seed21(int n, FLOAT *x)
/* Denote by \rho the distribution density of the points:
 *	\rho = 1 / Sqrt[\lamba_0 \lambda_1 \lambda_2]
 * Let \lambda_0 = 1-t/2, \lambda_1 = 1-t/2, \lambda_2 = t, then:
 *	\rho(t) = 1 / Sqrt[(1-t/2)^2 t]
 *	\int_0^x\rho(t) dt = Integrate[1 / Sqrt[(1-t/2)^2 t], {t, 0, x}]
 *			   = Sqrt[8] ArcTanh[Sqrt[x/2]]
 * Solve[Sqrt[8] ArcTanh[Sqrt[t/2]] == y, t] yields:
 *	t = 2 Tanh[y/Sqrt[8]]^2
 * Thus
 *	x = (1-t)/2 = 1/2 - Tanh[y/Sqrt[8]]^2
 */
{
    int i;
#if USE_INTERVALS
#if 0
    FLOAT t[n + 1], y0;
    y0 = Sqrt(8.) * Atanh(Sqrt((1./2.)));
    t[0] = 0.;
    t[n] = 0.5;
    for (i = 1; i < n; i++) {
	t[i] = 0.5 - Pow(Tanh((n - i) * y0 / (n * Sqrt(8.))), 2);
    }
    for (i = 0; i < n; i++, x++) {
	x[0] = t[i] + Random(t[i + 1] - t[i]);
printf("perm21(%lg)\n", (double)x[0]);
    }
#else
    static FLOAT *t = NULL, a;
    int j0, j1;
    if (t == NULL) {
	density_2D();
	t = malloc((2 * den_n + 1) * sizeof(*t));
	t[0] = 0.;
	for (i = 1; i <= den_n; i++) {
	    t[i] = t[i - 1] + den[i] - den[i - 1];
	}
	for (; i <= 2 * den_n; i++) {
	    t[i] = t[i - 1] + den[2 * den_n - i + 1] - den[2 * den_n - i];
	}
    }
    j0 = j1 = 0;
    for (i = 0; i < n; i++, x++) {
	a = (i + 1) * t[2 * den_n] / n;
	if (a > t[j1]) {
	    j0 = j1;
	    while (j1 < 2 * den_n && a > t[j1])
		j1++;
	}
	a = (j0 + Random(j1-j0)) / (2.L * den_n);
	if (a <= .5)
	    a = (.5 - a) * 2. / 3.L;
	else
	    a = 1. / 3.L + (1. - a) / 3.;
	x[0] = a;
//printf("perm21(%lg)\n", (double)x[0]);
    }
#endif 
    return;
#endif	/* USE_INTERVALS */

    for (i = 0; i < n; i++, x++) {
#if USE_EXTREMAL_MEASURE
	FLOAT a, f;
	while (1) {
	    a = Random(1.0);
	    f = Random(EXTREMAL_TRUNC);
	    if (a > .5)
		continue;
	    if (a == 0. || a == 0.5)
		break;
	    if (f < 1. / (a * Sqrt(1. - a - a)));
		break;
	}
	x[0] = a;
#else	/* USE_EXTREMAL_MEASURE */
	/* uniform distribution */
	x[0] = Random(0.5);
#endif	/* USE_EXTREMAL_MEASURE */
    }
}

static void
seed111(int n, FLOAT *x)
/* The triangle is decomposed into a sequence of nested triangles
 * sharing the same barycentric center. The triangles T(a), a>0 && a < 1/3,
 * is defined by: x >= 1/3-a && y >= 1/3-a && x+y <= 1/3+2a */
{
    int i;
#if USE_INTERVALS
    int j0, j1;
    FLOAT a;

    if (n == 0)
	return;
    density_2D();
    j0 = j1 = 0;
    for (i = 0; i < n; i++, x += 2) {
	a = (i + 1) * den[den_n] / n;
	if (a > den[j1]) {
	    j0 = j1;
	    while (j1 < den_n && a > den[j1])
		j1++;
	}
	x[0] = (1. - j1 / (FLOAT)den_n + Random((j1-j0) / (FLOAT)den_n)) / 3.L;
	x[1] = x[0] + Random(1. - 3. * x[0]);
//printf("perm111(%lg,%lg)\n", (double)x[0], (double)x[1]);
    }
#else	/* USE_INTERVALS */
    for (i = 0; i < n; i++, x += 2) {
#if USE_EXTREMAL_MEASURE
	FLOAT a, b, f;
	while (1) {
	    a = Random(1.0);
	    b = Random(1.0);
	    f = Random(EXTREMAL_TRUNC);
	    if (b >= a || b <= .5 * (1. - a) || a + b > 1)
		continue;
	    if (a == 1. || b == 0. || a + b == 1.)
		break;
	    if (f < 1. / Sqrt(a * b * (1. - a - b)));
		break;
	}
	x[0] = a;
	x[1] = b;
#else	/* USE_EXTREMAL_MEASURE */
	x[0] = Random(1.0);
	x[1] = Random(1.0 - x[0]);
#endif	/* USE_EXTREMAL_MEASURE */
    }
#endif	/* USE_INTERVALS */
}

static void
seed31(int n, FLOAT *x)
{
    int i;

    for (i = 0; i < n; i++, x++) {
#if USE_EXTREMAL_MEASURE
	FLOAT a, f;
	while (1) {
	    a = Random(1.0);
	    f = Random(EXTREMAL_TRUNC);
	    if (a > 1. / 3.L)
		continue;
	    if (a == 0. || Fabs(a - (FLOAT)(1./3.L)) < 1e10)
		break;
	    if (f < 1. / Sqrt(a * a * a * (1. - 3. * a)));
		break;
	}
	x[0] = a;
#else	/* USE_EXTREMAL_MEASURE */
	x[0] = Random(1. / 3.L);
#endif	/* USE_EXTREMAL_MEASURE */
    }
}

static void
seed22(int n, FLOAT *x)
{
    int i;

    for (i = 0; i < n; i++, x++) {
#if USE_EXTREMAL_MEASURE
	FLOAT a, f;
	while (1) {
	    a = Random(1.0);
	    f = Random(EXTREMAL_TRUNC);
	    if (a >= 0.25)
		continue;
	    if (a == 0.)
		break;
	    if (f < 1. / Sqrt(a * a * (.5 - a) * (.5 - a)));
		break;
	}
	x[0] = a;
#else	/* USE_EXTREMAL_MEASURE */
	x[0] = Random(1. / 4.L);
#endif	/* USE_EXTREMAL_MEASURE */
    }
}

static void
seed211(int n, FLOAT *x)
{
    int i;

    for (i = 0; i < n; i++, x += 2) {
#if USE_EXTREMAL_MEASURE
	FLOAT a, b, f;
	while (1) {
	    a = Random(1.0);
	    b = Random(1.0);
	    f = Random(EXTREMAL_TRUNC);
	    if (b == a || a >= .5 || a + a + b > 1. || b <= .5 * (1. - a - a))
		continue;
	    if (a == 0. || a == .5 || b == 0. || b == 1. || a + a + b == 1.)
		break;
	    if (f < 1. / Sqrt(a * a * b * (1. - a - a - b)));
		break;
	}
	x[0] = a;
	x[1] = b;
#else	/* USE_EXTREMAL_MEASURE */
	x[0] = Random(.5);
	x[1] = Random(1. - 2. * x[0]);
#endif	/* USE_EXTREMAL_MEASURE */
    }
}

static void
seed1111(int n, FLOAT *x)
{
    int i;

    for (i = 0; i < n; i++, x += 3) {
#if USE_EXTREMAL_MEASURE
	FLOAT a, b, c, f;
	while (1) {
	    a = Random(1.0);
	    b = Random(1.0);
	    c = Random(1.0);
	    f = Random(EXTREMAL_TRUNC);
	    if (a == 1. || b >= a || c >= b || a + b + c > 1.)
		continue;
	    if (c == 0. || a + b + c == 1.) 
		break;
	    if (f < 1. / Sqrt(a * b * c * (1. - a - b - c)));
		break;
	}
	x[0] = a;
	x[1] = b;
	x[2] = c;
#else	/* USE_EXTREMAL_MEASURE */
	x[0] = Random(1.);
	x[1] = Random(1. - x[0]);
	x[2] = Random(1. - x[0] - x[1]);
#endif	/* USE_EXTREMAL_MEASURE */
    }
}

static void
abscissas_seed(FLOAT D[])
/* this function generates an initial guess on the abscissas */
{
#if SYMMETRY
    int j;

    if (dim == 2) {
	seed21(s21, D);
	j = s21;
	seed111(s111, D + j);
	j += 2 * s111;
    }
    else {
	seed31(s31, D);
	j = s31;
	seed22(s22, D + j);
	j += s22;
	seed211(s211, D + j);
	j += 2 * s211;
	seed1111(s1111, D + j);
	j += 3 * s1111;
    }
    assert(j == nabsc);
#else
    for (i = 0; i < npoints; i += dim) {
	D[i] = Random(1.0);
	D[i + 1] = Random(1.0 - D[i]);
	if (dim == 3)
	    D[i + 2] = Random(1.0 - D[i] - D[i + 1]);
    }
#endif
}

static FLOAT
compute_error(const int wmap[], const FLOAT points[], const FLOAT weights[])
/* compute errors for the basis functions [neqns, npolys) */
{
    FLOAT error = 0., quad, e;
    const FLOAT *x;
    int i, j, n0, n1, nd;

    n0 = n1 = 0;
    for (i = 0; i < npolys; i++) {
	quad = 0.;
	/* compute numerical quadrature of the basis function i */
	for (j = 0, x = points; j < npoints; j++, x += dim) {
	    if ((e = weights[wmap[j]]) == 0.0)
		continue;
	    quad += e * poly_eval(pbasis[i], x[0], x[1],
					(dim == 2) ? (FLOAT)0.:x[2]);
	}
	e = quad - (!ORTHOGONAL_BASIS || i == 0 ? 1.0 : 0.0);
	if (Fabs(e) < EPS) {
	    if (n0 == i)
		n0++;
	    else
		n1++;
	}
	error += e * e;
    }

    for (nd = 0; nd <= degree && pdims[nd] <= n0; nd++) ;
    nd--;

    if (print) {
	printf("  Degree of the rule: %d\n", nd);
	printf("  Exactness of the rule: %d first + %d extra\n", n0, n1);
    }

    return Sqrt(error / npolys);
}

static FLOAT
compute_weights(const FLOAT D[], FLOAT x[][dim],
		FLOAT w[], FLOAT wjac[], int wmap[])
/* computes the weights for the given abscissas, and optional the Jacobian
 * with respect to the abscissas (if wjac != NULL).
 *
 * returns the L2 error with the integrals of all basis functions */
{
    POLY **basis;
    FLOAT *A, a;
    int i, j, k, M, N, NB;

#define A_(i,j)	A[(i) * (N + NB) + (j)]

#if USE_CARDINAL
if (dim == 2) {
    j = 0;
    for (i = 0; i < s21; i++) {
	if (Fabs(D[j] - 1.L/3.L) < 1e-10) {
	    fprintf(stderr, "WARNING: error 1\n");
	}
	j++;
    }
    for (i = 0; i < s111; i++) {
	if (Fabs(D[j] - D[j + 1]) < 1e-10) {
	    fprintf(stderr, "WARNING error 2\n");
	}
	j += 2;
    }
}
else {
    j = 0;
    for (i = 0; i < s31; i++) {
	if (Fabs(D[j] - .25) < 1e-10) {
	    fprintf(stderr, "WARNING error 1\n");
	}
	j++;
    }
    for (i = 0; i < s22; i++) {
	if (Fabs(D[j] - .25) < 1e-10) {
	    fprintf(stderr, "WARNING error 2\n");
	}
	j++;
    }
    for (i = 0; i < s211; i++) {
	if (Fabs(D[j] - D[j + 1]) < 1e-10 ||
	    Fabs(.5 * D[j] - D[j + 1]) < 1e-10) {
	    fprintf(stderr, "WARNING error 3\n");
	}
	j += 2;
    }
    for (i = 0; i < s1111; i++) {
	if (Fabs(D[j] - D[j + 1]) < 1e-10 ||
	    Fabs(D[j + 1] - D[j + 2]) < 1e-10 ||
	    Fabs(D[j] - D[j + 2]) < 1e-10) {
	    fprintf(stderr, "WARNING error 4\n");
	}
	j += 3;
    }
}

    /* use cardinal functions */
    basis = obasis;
    M = NCARD;
    N = npoints;
    NB = N;
#else	/* USE_CARDINAL */
    basis = pbasis;
    M = npolys;
    N = nwgts;
    NB = 1;
#endif	/* USE_CARDINAL */
    
    /* compute coefficients of the cardinal functions */

    A = malloc(M * (N + NB) * sizeof(A[0]));
    /* compute Newton-Cotes or least square quadrature weights */
    for (i = 0; i < M; i++) {
#if USE_CARDINAL
	for (j = 0; j < npoints; j++) {
	    A_(i, j) = poly_eval(basis[i], x[j][0], x[j][1],
					dim == 2 ? (FLOAT)0. : x[j][2]);
	}
	for (j = 0; j < NB; j++) {
	    A_(i, N + j) = (i == j) ? 1.0 : 0.0;
	}
#else	/* USE_CARDINAL */
	memset(&A_(i, 0), 0, (N + NB) * sizeof(*A));
	for (j = 0; j < npoints; j++) {
	    A_(i, wmap[j]) += poly_eval(basis[i], x[j][0], x[j][1],
					dim == 2 ? (FLOAT)0. : x[j][2]);
	}
#if ORTHOGONAL_BASIS
	A_(i, N) = (i == 0) ? 1.0 : 0.0;
#else	/* ORTHOGONAL_BASIS */
	A_(i, N) = (i == 0 || basis == pbasis) ? 1.0 : 0.0;
#endif	/* ORTHOGONAL_BASIS */
#endif	/* USE_CARDINAL */
    }

    k = linear_solver(M, N, NB, N + NB, (void *)A);

    if (k != 0) {
	if (k > 0) {
	    if (print)
	    fprintf(stderr, "WARNING: underdetermined lin. sys. (-%d).\n", k);
	}
	else {
	    fprintf(stderr, "ERROR: can't compute cardinal functions.\n");
#if USE_MPI
	    MPI_Abort(MPI_COMM_WORLD, 1);
#endif
	    exit(1);
	}
    }

    /* compute the weights */
    a = 0.;
    for (j = 0; j < npoints; j++) {
#if USE_CARDINAL
	w[wmap[j]] = A_(j, N);
#else	/* USE_CARDINAL */
	w[wmap[j]] = A_(wmap[j], N);
#endif	/* USE_CARDINAL */
	if (w[wmap[j]] == 0.0)
	    continue;
	if (print) {
	    if (dim == 2)
		printf(" %13.6le %13.6le %13.6le, %13.6le\n", (double)x[j][0],
			(double)x[j][1], (double)(1. - x[j][0] - x[j][1]),
			(double)w[wmap[j]]);
	    else
		printf(" %13.6le %13.6le %13.6le %13.6le, %13.6le\n",
			(double)x[j][0], (double)x[j][1], (double)x[j][2],
			(double)(1. - x[j][0] - x[j][1] - x[j][2]),
			(double)w[wmap[j]]);
	}
    }

    a = compute_error(wmap, (void *)x, w);

    free(A);

#undef A_
    return a;
}

#define OUTPUT_VTK	1	/* == 0 ==> output gnuplot format */

#if OUTPUT_VTK
#define VTK_OUT(value, var)						\
	var = value;							\
	fwrite(swap_endian(&var, sizeof(var)), sizeof(var), 1, f);

static void *
swap_endian(const void *p0, size_t size)
{
    static unsigned char s[16];
    const unsigned char *p;
    unsigned char *q;

    p = p0;
    q = s + size;

    assert(sizeof(s) >= size);
    while (q > s) *(--q) = *(p++);

    return s;
}
#endif

static void
plot_gen(int n, int n0, FLOAT D0, FLOAT D[])
{
    static int no = 0;
    static char fn[PATH_MAX + 1];
    FILE *f;
    int i, j, k, N, L;
    FLOAT err, min, max, d;
    int *wmap;
    FLOAT (*x)[dim], *w;
#if OUTPUT_VTK
    float fv;
    int   iv;
#endif

    if (n <= 0 || n > 3) {
	fprintf(stderr, "plot_gen: dimension != {1,2,3}, case skipped!\n");
	return;
    }

    x = malloc(npoints * sizeof(x[0]));
    wmap = malloc(npoints * sizeof(*wmap));
    w = malloc(nwgts * sizeof(*w));
    if (dim == 2)
	abscissas_2D(D, x, wmap);
    else
	abscissas_3D(D, x, wmap);

    sprintf(fn, "quad%d.%s", no++, OUTPUT_VTK && n > 1 ? "vtk" : "dat");
    f = fopen(fn, "w+t");

    if (n == 1) {
	N = 1000;
	d = D0 / N;
	min = 1e10;
	max = 0.;
	for (i = 0; i <= N; i++) {
	    D[n0] = i * d;
	    err = compute_weights(D, x, w, NULL, wmap);
	    if (min > err)
		min = err;
	    if (max < err)
		max = err;
	    /* err = (err > 1.0) ? 1.0 : err */
	    fprintf(f, "%lg %lg\n", (double)D[n0], (double)err);
	    fprintf(stderr, "i=%-4d\r", i);
	    fflush(stdout);
	    fflush(stderr);
	}
	fprintf(stderr, "min: %lg, max: %lg\n", (double)min, (double)max);
	fprintf(stderr, "echo 'plot \"%s\"; pause 1000' | gnuplot\n", fn);
    }
    else if (n == 2) {
	N = 100;
	d = D0 / N;
	min = 1e10;
	max = 0.;
#if OUTPUT_VTK
	fprintf(f, "# vtk DataFile Version 2.0\n\n"
		   "BINARY\nDATASET UNSTRUCTURED_GRID\n"
		   "POINTS %d float\n", (N + 1) * (N + 1));
	for (i = 0; i <= N; i++) {
	    for (j = 0; j <= N; j++) {
		VTK_OUT(i * d, fv); VTK_OUT(j * d, fv); VTK_OUT((FLOAT)0., fv);
	    }
	}
	fprintf(f, "CELLS %d %d\n", N * N * 2, N * N * 2 * 4);
	L = 0;
	for (i = 0; i < N; i++) {
	    for (j = 0; j < N; j++, L++) {
		VTK_OUT(3, iv);
		VTK_OUT(L, iv);
		VTK_OUT(L + 1, iv);
		VTK_OUT(L + N+1, iv);
		VTK_OUT(3, iv);
		VTK_OUT(L + 1 + N+1, iv);
		VTK_OUT(L + 1, iv);
		VTK_OUT(L + N+1, iv);
	    }
	    L++;
	}
	fprintf(f, "CELL_TYPES %d\n", N * N * 2);
	for (i = 0; i < N; i++) {
	    for (j = 0; j < N; j++) {
		VTK_OUT(5, iv); VTK_OUT(5, iv);
	    }
	}
	fprintf(f, "POINT_DATA %d\n"
		   "SCALARS error float 1\n"
		   "LOOKUP_TABLE default\n", (N + 1) * (N + 1));
#endif
	for (i = 0; i <= N; i++) {
	    D[n0] = i * d;
	    for (j = 0; j <= N; j++) {
		D[n0 + 1] = j * d;
		err = compute_weights(D, x, w, NULL, wmap);
		if (min > err)
		    min = err;
		if (max < err)
		    max = err;
		/* err = (err > 1.0) ? 1.0 : err */
#if OUTPUT_VTK
		VTK_OUT(err, fv);
#else
		fprintf(f, "%lg %lg %lg\n", (double)D[n0], (double)D[n0 + 1],
				(double)err);
#endif
		fprintf(stderr, "i=%-4d j=%-4d\r", i, j);
		fflush(stderr);
	    }
#if !OUTPUT_VTK
	    fprintf(f, "\n");
#endif
	}
	fprintf(stderr, "min: %lg, max: %lg\n", (double)min, (double)max);
#if OUTPUT_VTK
	fprintf(stderr, "\"%s\" created.\n", fn);
#else
	fprintf(stderr, "echo 'set pm3d; set mouse; splot \"%s\" with "
			"pm3d; pause mouse' | gnuplot\n", fn);
#endif
    }
    else {
	N = 20;
	d = D0 / N;
	min = 1e10;
	max = 0.;
#if OUTPUT_VTK
	fprintf(f, "# vtk DataFile Version 2.0\n\n"
		   "BINARY\nDATASET UNSTRUCTURED_GRID\n"
		   "POINTS %d float\n", (N + 1) * (N + 1) * (N + 1));
	for (i = 0; i <= N; i++) {
	    for (j = 0; j <= N; j++) {
		for (k = 0; k <= N; k++) {
		    VTK_OUT(i * d, fv);
		    VTK_OUT(j * d, fv);
		    VTK_OUT(k * d, fv);
		}
	    }
	}
	fprintf(f, "CELLS %d %d\n", N * N * N * 5, N * N * N * 5 * 5);
	L = 0;
	for (i = 0; i < N; i++) {
	    for (j = 0; j < N; j++) {
		for (k = 0; k < N; k++, L++) {
#define I(a,b,c)	L + (N + 1) * ((N + 1) * (a) + (b)) + (c)
		    VTK_OUT(4, iv);
		    VTK_OUT(I(0,0,0), iv); VTK_OUT(I(1,0,0), iv);
		    VTK_OUT(I(0,1,0), iv); VTK_OUT(I(0,0,1), iv);
		    VTK_OUT(4, iv);
		    VTK_OUT(I(1,0,0), iv); VTK_OUT(I(0,1,0), iv);
		    VTK_OUT(I(0,0,1), iv); VTK_OUT(I(1,1,1), iv);
		    VTK_OUT(4, iv);
		    VTK_OUT(I(1,0,0), iv); VTK_OUT(I(0,1,0), iv);
		    VTK_OUT(I(1,1,0), iv); VTK_OUT(I(1,1,1), iv);
		    VTK_OUT(4, iv);
		    VTK_OUT(I(1,0,0), iv); VTK_OUT(I(0,0,1), iv);
		    VTK_OUT(I(1,0,1), iv); VTK_OUT(I(1,1,1), iv);
		    VTK_OUT(4, iv);
		    VTK_OUT(I(0,1,0), iv); VTK_OUT(I(0,0,1), iv);
		    VTK_OUT(I(0,1,1), iv); VTK_OUT(I(1,1,1), iv);
#undef I
	        }
		L++;
	    }
	    L += N + 1;
	}
	fprintf(f, "CELL_TYPES %d\n", N * N * N * 5);
	for (i = 0; i < N; i++) {
	    for (j = 0; j < N; j++) {
		for (k = 0; k < N; k++) {
		    VTK_OUT(10, iv); VTK_OUT(10, iv); VTK_OUT(10, iv);
		    VTK_OUT(10, iv); VTK_OUT(10, iv);
		}
	    }
	}
	fprintf(f, "POINT_DATA %d\n"
		   "SCALARS error float 1\n"
		   "LOOKUP_TABLE default\n", (N + 1) * (N + 1) * (N + 1));
#endif
	for (i = 0; i <= N; i++) {
	    D[n0] = i * d;
	    for (j = 0; j <= N; j++) {
		D[n0 + 1] = j * d;
		for (k = 0; k <= N; k++) {
		    D[n0 + 2] = 0.2 /*k / 100.0*/;
		    err = compute_weights(D, x, w, NULL, wmap);
		    if (min > err)
			min = err;
		    if (max < err)
			max = err;
		    /* err = (err > 1.0) ? 1.0 : err */
#if OUTPUT_VTK
		    VTK_OUT(err, fv);
#else
		    fprintf(f, "%lg %lg %lg\n", (double)D[n0],
					(double)D[n0 + 1], (double)err);
#endif
		    fprintf(stderr, "i=%-4d j=%-4d\r", i, j);
		    fflush(stderr);
		}
	    }
#if !OUTPUT_VTK
	    fprintf(f, "\n");
#endif
	}
	fprintf(stderr, "min: %lg, max: %lg\n", (double)min, (double)max);
#if OUTPUT_VTK
	fprintf(stderr, "\"%s\" created.\n", fn);
#else
	fprintf(stderr, "echo 'set pm3d; set mouse; splot \"%s\" with "
			"pm3d; pause mouse' | gnuplot\n", fn);
#endif
    }
    fclose(f);

    free(x);
    free(wmap);
    free(w);

    return;
}

typedef void (*FCN_TYPE0)(int *M, int *N, double X[], double FVEC[],
	double FJAC[], int *LDFJAC, int *IFLAG);

static int eval_count = 0;

static void
func0(int *M, int *N, double X0[], double FVEC[], double FJAC[],
      int *LDFJAC, int *IFLAG)
/* callback function for MINPACK's subroutine LMDER1(), which computes
 * the integration errors and the Jacobian */
{
    int i, j, l0, l1, *wmap;
    FLOAT (*x)[dim], (*xjac)[dim], *w, d, e, f;
    POLY *p;
#if USE_LONG_DOUBLE
    FLOAT *X;
#else	/* USE_LONG_DOUBLE */
    FLOAT *X = X0;
#endif	/* USE_LONG_DOUBLE */

#if ASYNCHRONOUS
    static time_t tcurr, tlast = 0;

    if ((tcurr = time(NULL)) - tlast >= 3) {
	tlast = tcurr;
	if (sync_file(CHECK_SYNC_FILE))
	    longjmp(jmp, 1);
    }
#endif	/* ASYNCHRONOUS */

#if USE_LONG_DOUBLE
    X = malloc((*N) * sizeof(*X));
    for (i = 0; i < *N; i++)
	X[i] = X0[i];
#endif	/* USE_LONG_DOUBLE */

    eval_count++;

    assert(*M == neqns && (*N == nabsc + nwgts || *N == nabsc));

    x = malloc(npoints * sizeof(x[0]));
    wmap = malloc(npoints * sizeof(*wmap));
    /*if (FJAC != NULL) -- to avoid stupid warning of GCC4 */
        xjac = malloc(1 * npoints * sizeof(xjac[0]));
    if (*N == nabsc) {
	/* set w using linear least square solution */
	w = malloc(nwgts * sizeof(*w));
	if (dim == 2)
	    abscissas_2D(X, x, wmap);
	else
	    abscissas_3D(X, x, wmap);
	compute_weights(X, x, w, NULL, wmap);
    }
    else {
	w = X + nabsc;
	if (dim == 2)
	    abscissas_2D(X, x, wmap);
	else
	    abscissas_3D(X, x, wmap);
    }

    for (i = 0; i < (*M); i++) {
	l0 = 0;
	f = d = 0.;
	for (j = 0; j < npoints; j++) {
	    l1 = wmap[j];
	    if (l0 != l1) {
		f += d * w[l0];
		if (*N > nabsc && FJAC != NULL)
		    FJAC[(nabsc + l0) * (*LDFJAC) + i] = d;
		else
		    (void)d;	/* TODO */
		l0 = l1;
		d = 0.;
	    }
	    if (dim == 2) {
		d += poly_eval(p = pbasis[i], x[j][0], x[j][1], (FLOAT)0.);
		if (FJAC != NULL) {
		    xjac[j][0] = poly_eval_d(p,0,x[j][0],x[j][1],(FLOAT)0.) * w[l0];
		    xjac[j][1] = poly_eval_d(p,1,x[j][0],x[j][1],(FLOAT)0.) * w[l0];
		}
	    }
	    else {
		d += poly_eval(p = pbasis[i], x[j][0], x[j][1], x[j][2]);
		if (FJAC != NULL) {
		    xjac[j][0] = poly_eval_d(p, 0, x[j][0],x[j][1],x[j][2])
				* w[l0];
		    xjac[j][1] = poly_eval_d(p, 1, x[j][0],x[j][1],x[j][2])
				* w[l0];
		    xjac[j][2] = poly_eval_d(p, 2, x[j][0],x[j][1],x[j][2])
				* w[l0];
		}
	    }
	}
	if (FVEC != NULL)
	    FVEC[i] = f + d * w[l0] - (!ORTHOGONAL_BASIS || i == 0 ? 1. : 0.);
	if (*N > nabsc && FJAC != NULL) {
	    FJAC[(nabsc + l0) * (*LDFJAC) + i] = d;
	    /* compute FJAC[0 ... nabsc - 1] */
	    if (dim == 2) {
		/* Note: should match with the order in abscissas_2D */
		j = s3;
		l0 = 0;
		for (l1 = 0; l1 < s21; l1++) {
		    FJAC[(l0++) * (*LDFJAC) + i] =
				+    xjac[j+0][0] +    xjac[j+0][1]
				+    xjac[j+1][0] - 2.*xjac[j+1][1]
				- 2.*xjac[j+2][0] +    xjac[j+2][1];
		    j += 3;
		}
		for (l1 = 0; l1 < s111; l1++) {
		    FJAC[(l0++) * (*LDFJAC) + i] =
				+ xjac[j+0][0] + xjac[j+1][1]
				+ xjac[j+2][0] - xjac[j+2][1]
				- xjac[j+3][0] + xjac[j+3][1]
				- xjac[j+4][1] - xjac[j+5][0];
		    FJAC[(l0++) * (*LDFJAC) + i] =
				+ xjac[j+0][1] + xjac[j+1][0]
				- xjac[j+2][1] - xjac[j+3][0]
				+ xjac[j+4][0] - xjac[j+4][1]
				- xjac[j+5][0] + xjac[j+5][1];
		    j += 6;
		}
		assert(j = npoints);
	    }
	    else {	/* dim == 3 */
/* Note: The macros Permxxx must match those in abscissas_3D() */
#define Perm21(A, B) 						\
	A xjac[j + 0][0] A xjac[j + 0][1] B xjac[j + 0][2]	\
	A xjac[j + 1][0] B xjac[j + 1][1] A xjac[j + 1][2]	\
	B xjac[j + 2][0] A xjac[j + 2][1] A xjac[j + 2][2]
#define Perm111(A, B, C)					\
	A xjac[j + 0][0] B xjac[j + 0][1] C xjac[j + 0][2]	\
	A xjac[j + 1][0] C xjac[j + 1][1] B xjac[j + 1][2]	\
	B xjac[j + 2][0] C xjac[j + 2][1] A xjac[j + 2][2]	\
	B xjac[j + 3][0] A xjac[j + 3][1] C xjac[j + 3][2]	\
	C xjac[j + 4][0] A xjac[j + 4][1] B xjac[j + 4][2]	\
	C xjac[j + 5][0] B xjac[j + 5][1] A xjac[j + 5][2]
		j = s4;
		l0 = 0;
		for (l1 = 0; l1 < s31; l1++) {
		    d = xjac[j][0] + xjac[j][1] + xjac[j][2];
		    j += 1;
		    FJAC[(l0++) * (*LDFJAC) + i] = d Perm21(+, -3.*);
		    j += 3;
		}
		for (l1 = 0; l1 < s22; l1++) {
		    d = Perm21(+, - .5*);
		    j += 3;
		    FJAC[(l0++) * (*LDFJAC) + i] = d Perm21(-.5*, +);
		    j += 3;
		}
		for (l1 = 0; l1 < s211; l1++) {
		    d = Perm21(+, +0.*);
		    e = Perm21(+0.*, +);
		    j += 3;
		    d += Perm21(+, -2.*);
		    e += Perm21(+0.*, -);
		    j += 3;
		    FJAC[(l0++) * (*LDFJAC) + i] = d Perm111(+, +0.*, -2.*);
		    FJAC[(l0++) * (*LDFJAC) + i] = e Perm111(+0.*, +, -);
		    j += 6;
		}
		for (l1 = 0; l1 < s1111; l1++) {
		    d = Perm111(+, +0.*, +0.*);
		    e = Perm111(+0.*, +, +0.*);
		    f = Perm111(+0.*, +0.*, +);
		    j += 6;
		    d += Perm111(+, +0.*, -);
		    e += Perm111(+0.*, +, -);
		    f += Perm111(+0.*, +0.*, -);
		    j += 6;
		    d += Perm111(+, +0.*, -);
		    e += Perm111(+0.*, +0.*, -);
		    f += Perm111(+0.*, +, -);
		    j += 6;
		    FJAC[(l0++) * (*LDFJAC) + i] = d Perm111(+0.*, +0.*, -);
		    FJAC[(l0++) * (*LDFJAC) + i] = e Perm111(+, +0.*, -);
		    FJAC[(l0++) * (*LDFJAC) + i] = f Perm111(+0.*, +, -);
		    j += 6;
		}
		assert(j = npoints);
#undef Perm21
#undef Perm111
	    }
	}
	else {
	    /* TODO: process Dw / Dx */
	}
    }

#if 0
d = 0.;
for (i = 0; i < *M; i++)
    if (d < Fabs(FVEC[i]))
	d = Fabs(FVEC[i]);
fprintf(stderr, "\t%d: error = %lg\n", eval_count, (double)d);
#endif

    if (xjac != NULL)
	free(xjac);
    free(x);
    free(wmap);
    if (*N == nabsc)
	free(w);
#if USE_LONG_DOUBLE
    free(X);
#endif	/* USE_LONG_DOUBLE */
}

typedef void (*FCN_TYPE)(int *M, int *N, double X[], double FVEC[], int *IFLAG);

static void
func(int *M, int *N, double X[], double FVEC[], int *IFLAG)
{
    func0(M, N, X, FVEC, NULL, NULL, IFLAG);
}

/*========================== Genetic Algorithm =========================*/

#if GENETIC

#include "gaul.h"

static boolean
ga_score(population *pop, entity *entity)
{
    int i, m = neqns, n = nabsc;
    double *FVEC, *X = (double *)entity->chromosome[0], error, d;

    for (i = 0; i < n; i++) {
	if (X[i] < 0. || X[i] > 1.) {
	    entity->fitness = -100000.0;
	    return TRUE;
	}
    }

    FVEC = malloc(m * sizeof(*FVEC));
    func(&m, &n, X, FVEC, &i);    

    error = 0.;
    for (i = 0; i < m; i++) {
	if (error < (d = Fabs(FVEC[i])))
	    error = d;
    }

    entity->fitness = - error;
    free(FVEC);

    return TRUE;
}

static boolean
ga_callback(int generation, population *pop)
{
    double error = - /*pop->entity_iarray[0]->fitness*/
		     ga_get_entity_from_rank(pop,0)->fitness;

    printf("%d: fitness = %lg\n", generation, (double)error);

    return error < REPORT_TOL ? FALSE : TRUE;
}

static boolean
ga_seed(population *pop, entity *adam)
{
    if (!pop)
	die("Null pointer to population structure passed.");
    if (!adam)
	die("Null pointer to entity structure passed.");

    /* Seeding */
    abscissas_seed((double *)adam->chromosome[0]);

    return TRUE;
}

static void
genetic(double D[])
{
    static int seeded = 0;
    population *pop;
    int nevol;

    if (nabsc == 0)
	return;

    if (!seeded) {
	seeded = 1;
	random_seed(getpid());
    }

    pop = ga_genesis_double(
	1000,				/* const int population_size */
       	1,				/* const int num_chromo */
       	nabsc,				/* const int len_chromo */
       	ga_callback,			/* GAgeneration_hook generation_hook */
       	NULL,				/* GAiteration_hook iteration_hook */
       	NULL,				/* GAdata_destructor data_destructor */
       	NULL,				/* GAdata_ref_incrementor
						data_ref_incrementor*/
       	ga_score,			/* GAevaluate evaluate */
       	ga_seed,			/* GAseed seed */
       	NULL,				/* GAadapt adapt */
       	ga_select_one_bestof2,		/* GAselect_one select_one */
       	ga_select_two_bestof2,		/* GAselect_two select_two */
       	ga_mutate_double_singlepoint_drift,	/* GAmutate mutate */
       	ga_crossover_double_doublepoints,	/* GAcrossover crossover */
       	NULL,				/* GAreplace replace */
       	NULL				/* vpointer User data */
    );

    ga_population_set_parameters(
	pop,				/* population *pop */
	GA_SCHEME_DARWIN,		/* const ga_scheme_type scheme */
	GA_ELITISM_PARENTS_SURVIVE,	/* const ga_elitism_type elitism */
	0.8,				/* double crossover */
	0.2,				/* double mutation */
	0.0      		        /* double migration */
    );

    if (TRUE) {
	/* Genetic Algorithm */
    	nevol = ga_evolution(pop, 5000);
    }
    else {
	/* Simulated anealing */
	entity *solution;
	ga_population_set_sa_parameters(
	    pop,			/* population *pop */
	    ga_sa_boltzmann_acceptance,	/* GAsa_accept SA acceptance criterion*/
	    100.0,			/* const double	Initial temperature */
	    0.0,		/* const double	Final temperature */
	    10.0,		/* const double	Temperature step size */
	    50			/* const int Temperature update frequency */
	);
	solution = ga_get_entity_from_rank(pop,0);
	nevol = ga_sa(pop, solution, 2000);
    }

fprintf(stderr, "GA stops after %d evolutions, fitness = %le.\n", nevol,
	(double)ga_get_entity_from_rank(pop,0)->fitness);
    memcpy(D, ga_get_entity_from_rank(pop,0)->chromosome[0],
		nabsc * sizeof(*D));

    ga_extinction(pop);
}

#endif	/* GENETIC */

static void
init_w(int n, FLOAT X[])
{
    if (n > nabsc) {
#if 1
	FLOAT (*x)[dim];
	int *wmap;

	/* compute weights */
	x = malloc(npoints * sizeof(x[0]));
	wmap = malloc(npoints * sizeof(wmap[0]));
	if (dim == 2)
	    abscissas_2D(X, x, wmap);
	else
	    abscissas_3D(X, x, wmap);
	compute_weights(X, x, X + nabsc, NULL, wmap);
	free(wmap);
	free(x);
#else
	int i;
	FLOAT d = 0.;
        for (i = nabsc; i < n - 1; i++)
	    d += (X[i] = Random(1.0));
	X[i] = 1.0 - d;
#endif
    }
}

/*========================== End Genetic Algorithm =========================*/

#if USE_TENSOLVE

typedef void (*FCN_TYPE_F)(double X[], double F[], int *M, int *N);
typedef void (*FCN_TYPE_JAC)(double X[], double J[/*n*/]/*[maxm]*/,
		int *MAXM, int *M, int *N);

static void
func_f(double X[], double F[], int *M, int *N)
{
    int info;

    func(M, N, X, F, &info);
}

static void
func_jac(double X[], double JAC[], int *MAXM, int *M, int *N)
{
    int info;

    func0(M, N, X, NULL, JAC, MAXM, &info);
}

static FLOAT
nonlinear_lss(FLOAT D[], FLOAT W[])
{
    extern void tsdflt_(int *M, int *N, int *ITNLIM, int *JACFLG,
			double *GRADTL, double *STEPTL, double *FTOL,
			int *METHOD, int *GLOBAL, double *STEPMX, double *DLT,
			double TYPX[], double TYPF[], int *IPR, int *MSG);
    extern void tsneci_(int *MAXM, int *MAXN, int *MAXP, double X0[],
			int *M, int *N, double TYPX[], double TYPF[],
			int *ITNLIM, int *JACFLG, double *GRADTL,
			double *STEPTL, double *FTOL, int *METHOD, int *GLOBAL,
			double *STEPMX, double *DLT, int *IPR,
			double WRKUNC[], int *LUNC, double WRKNEM[], int *LNEM,
			double WRKNEN[], int *LNEN, int IWRKN[], int *LIN,
			FCN_TYPE_F FVEC, FCN_TYPE_JAC JAC, int *MSG,
			double XP[], double FP[], double GP[], int *TERMCD);
    /*extern FCN_TYPE_JAC tsdumj_;*/
    int		maxm, maxn, maxp;
    int		i, m, n, itnlim, jacflg, method,
		global, ipr, lunc, lnem, lnen, lin = 3, msg, termcd;
    double	gradtl, steptl, ftol, stepmx, dlt;
    int		*iwrkn;
    double	*x0, *xp, *fp, *gp, *typx, *typf, *wrknen, *wrkunc, *wrknem;
    double	error;

    m = neqns;
#if INDEPENDENT_W
    n = nabsc + nwgts;
#else
    n = nabsc;
#endif

    if (n == 0)
	n += nwgts;

    i = (int)(Sqrt(n) + 0.500000001);
    maxn = n + 2;
    maxm = maxn + m;
    maxp = i   + 1;
    lunc = 2 * i + 4   + 1;
    lnem = n + 2 * i + 11   + 1;
    lnen = 2 * i + 9   + 1;

    iwrkn = malloc(maxn * lin * sizeof(*iwrkn));
    x0 = malloc(maxn * sizeof(*x0));
    xp = malloc(maxn * sizeof(*xp));
    fp = malloc(maxm * sizeof(*fp));
    gp = malloc(maxn * sizeof(*gp));
    typx = malloc(maxn * sizeof(*typx));
    typf = malloc(maxm * sizeof(*typf));
    wrknen = malloc(maxn * lnen * sizeof(*wrknen));
    wrkunc = malloc(maxp * lunc * sizeof(*wrkunc));
    wrknem = malloc(maxm * lnem * sizeof(*wrknem));

    tsdflt_(&m, &n, &itnlim, &jacflg, &gradtl, &steptl, &ftol, &method,
	    &global, &stepmx, &dlt, typx, typf, &ipr, &msg);

    init_w(n, x0);

    jacflg = 1;
    global = 1;
    stepmx = 1.0;
    dlt = -1.0;

    gradtl = 1.0e-16;
    ftol   = 1.0e-16;
    steptl = 1.0e-16;
    global = 1;

    tsneci_(&maxm, &maxn, &maxp, x0, &m, &n, typx, typf, &itnlim, &jacflg,
	    &gradtl, &steptl, &ftol, &method, &global, &stepmx, &dlt, &ipr,
	    wrkunc, &lunc, wrknem, &lnem, wrknen, &lnen, iwrkn, &lin,
	    func_f, func_jac/*tsdumj_*/, &msg, xp, fp, gp, &termcd);

    error = 0.;
    for (i = 0; i < m; i++)
	if (Fabs(fp[i]) > error)
	    error = Fabs(fp[i]);
    //printf("    info = %d, error = %lg\n", info, (double)error);
    if (nabsc > 0) {
	for (i = 0; i < nabsc; i++)
	    D[i] = xp[i];
    }

    if (nabsc < n) {
	for (i = nabsc; i < n; i++) {
	    W[i - nabsc] = xp[i] * (dim == 2 ? 2. : 6.);
	}
    }
    else {
	FLOAT (*x)[dim];
	int *wmap;
	x = malloc(npoints * sizeof(x[0]));
	wmap = malloc(npoints * sizeof(wmap[0]));
	if (dim == 2)
	    abscissas_2D(D, x, wmap);
	else
	    abscissas_3D(D, x, wmap);
	compute_weights(D, x, W, NULL, wmap);
	for (i = 0; i < nwgts; i++)
	    W[i] *= (dim == 2 ? 2. : 6.);
	free(x);
	free(wmap);
    }

    free(iwrkn);
    free(x0);
    free(xp);
    free(fp);
    free(gp);
    free(typx);
    free(typf);
    free(wrknen);
    free(wrkunc);
    free(wrknem);

    return error;
}

#else	/* USE_TENSOLVE */

void
lmdif1parms_(int *maxfev, int *nprint)
/* callback function for setting some parameters for lmdif() */
{
    *nprint = 0;
    *maxfev = LMDIF_MAXFEV * (nabsc + 1);
}

static FLOAT
nonlinear_lss(FLOAT D[], FLOAT W[])
/* call MINPACK LMDER1 to find least square solutions of the current case */
{
    extern void lmder1_(FCN_TYPE0 fcn, int *m, int *n, double x[],
			double fvec[], double fjac[], int *ldfjac,
			double *tol, int *info, int ipvt[],
			double wa[], int *lwa);
    extern void lmdif1_(FCN_TYPE fcn, int *m, int *n, double x[],
			double fvec[], double *tol, int *info,
			int iwa[], double wa[], int *lwa);
    int m, n, ldfjac, info, lwa, *IPVT;
    double *X, *FVEC, *FJAC, *WA, tol, error;
    int i;

#if INDEPENDENT_W
    n = nabsc + nwgts;
#else
    n = nabsc;
#endif

    if (n == 0)
	n += nwgts;

    ldfjac = m = neqns;
    lwa = 5 * n + m + n * m;

    X = malloc(n * sizeof(*X));
    FVEC = malloc(m * sizeof(*FVEC));
    FJAC = malloc(m * n * sizeof(*FJAC));
    IPVT = malloc(n * sizeof(*IPVT));
    WA = malloc(lwa * sizeof(*WA));

#if USE_LONG_DOUBLE
  {
    FLOAT *X0 = malloc(n * sizeof(*X0));
    if (nabsc > 0)
	memcpy(X0, D, nabsc * sizeof(*X0));
    init_w(n, X0);
    for (i = 0; i < n; i++)
	X[i] = X0[i];
    free(X0);
  }
#else	/* USE_LONG_DOUBLE */
    if (nabsc > 0)
	memcpy(X, D, nabsc * sizeof(*X));
    init_w(n, X);
#endif	/* USE_LONG_DOUBLE */

    tol = EPS;
#if 0
    /* Note: even when analytic Jacobian is available, lmdif1_ seems to be
     * much more robust, though much slower than lmder1 */
    if (n > nabsc)
	lmder1_(func0, &m, &n, X, FVEC, FJAC, &ldfjac, &tol, &info,
		IPVT, WA, &lwa);
    else
#endif
	lmdif1_(func, &m, &n, X, FVEC, &tol, &info, IPVT, WA, &lwa);

    error = 0.;
    for (i = 0; i < m; i++) {
//printf("%d: order = %d + %d = %d, res = %lg\n", i, porders[2 * i], porders[2 * i + 1], porders[2 * i] + porders[2 * i + 1], (double)FVEC[i]);
	if (Fabs(FVEC[i]) > error)
	    error = Fabs(FVEC[i]);
    }
//printf("    info = %d, error = %lg\n", info, (double)error);
    if (nabsc > 0) {
	for (i = 0; i < nabsc; i++)
	    D[i] = X[i];
    }
    if (nabsc < n) {
	for (i = nabsc; i < n; i++) {
	    W[i - nabsc] = X[i] * (dim == 2 ? 2. : 6.);
	}
    }
    else {
	FLOAT (*x)[dim];
	int *wmap;
	x = malloc(npoints * sizeof(x[0]));
	wmap = malloc(npoints * sizeof(wmap[0]));
	if (dim == 2)
	    abscissas_2D(D, x, wmap);
	else
	    abscissas_3D(D, x, wmap);
	compute_weights(D, x, W, NULL, wmap);
	for (i = 0; i < nwgts; i++)
	    W[i] *= (dim == 2 ? 2. : 6.);
	free(x);
	free(wmap);
    }

    free(X);
    free(FVEC);
    free(FJAC);
    free(IPVT);
    free(WA);

    return error;
}

#endif	/* USE_TENSOLVE */

static int tmp_size;

static int
absc_comp(const void *p0, const void *p1)
{
    int i;
    const FLOAT *a0 = p0, *a1 = p1;

    for (i = 0; i < tmp_size; i++) {
	if (*a0 > *a1)
	    return 1;
	if (*a0 < *a1)
	    return -1;
    }

    return 0;
}

static void
sort_abscissas(int nlen, const int len[], const int size[], FLOAT *D)
{
    int i;

    for (i = 0; i < nlen; i++) {
	tmp_size = size[i];
	qsort(D, len[i], tmp_size * sizeof(*D), absc_comp);
	D += len[i] * size[i];
    }
}

static int
check_last_found(const FLOAT D[])
/* checks the current solution against the one found in the last try.
 * returns the number of identical solutions found for the current case */
{
    static int count = 0;
    static int t3 = -1, t21 = -1, t111 = -1;
    static int n2[] = {1, 1, 2};
    static int t4 = -1, t31 = -1, t22 = -1, t211 = -1, t1111 = -1;
    static int n3[] = {1, 1, 1, 2, 3};
    static FLOAT *D0 = NULL;
    FLOAT *D1;
    int i, flag, n, nlen, *n0, len[4];

    if (dim == 2) {
	nlen = 2;
	n = s21 + 2 * s111;
	len[0] = s21; len[1] = s111;
	n0 = n2;
	if (t3 != s3 || t21 != s21 || t111 != s111) {
	    flag = 0;
	    t3 = s3; t21 = s21; t111 = s111;
	}
	else {
	    flag = 1;
	}
    }
    else {
	nlen = 4;
	n = s31 + s22 + 2 * s211 + 3 * s1111;
	len[0] = s31; len[1] = s22; len[2] = s211; len[3] = s1111;
	n0 = n3;
	if (t4 != s4 || t31 != s31 || t22 != s22 || t211 != s211 ||
	    t1111 != s1111) {
	    flag = 0;
	    t4 = s4; t31 = s31; t22 = s22; t211 = s211; t1111 = s1111;
	}
	else {
	    flag = 1;
	}
    }

    if (n > 0) {
	D1 = malloc(n * sizeof(*D1));
	memcpy(D1, D, n * sizeof(*D0));
	sort_abscissas(nlen, len, n0, D1);
    }
    else {
	D1 = NULL;
    }

    if (flag) {
	assert(D0 != NULL);
	for (i = 0; i < n; i++)
#if USE_LONG_DOUBLE
	    if (Fabs(D0[i] - D1[i]) >= 1e-15)
		break;
#else	/* USE_LONG_DOUBLE */
	    if (Fabs(D0[i] - D1[i]) >= 1e-12)
		break;
#endif
	if (i >= n) {
	    return ++count;
	}
    }

    /* save the new solution */
    if (D0 != NULL)
	free(D0);
    D0 = D1;

    return (count = 0);
}

static int
check_found(int abort)
{
#if USE_MPI
    int n[2], m[2];
    n[0] = nsols;
    n[1] = abort;
    MPI_Allreduce(n, m, 2, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
    nsols = m[0];
    abort = m[1];
#endif
    return nsols != 0 || abort != 0;
}

static int
check_solution_2d(int prtflag, int tries, FLOAT error, FLOAT tol,
			FLOAT D[], FLOAT W[])
{
    int inbound = 1, i, j, k;

#if !SYMMETRY
# error unimplemented!
#endif

    if (prtflag) {
	time_t t = time(NULL);
	printf("** %s", ctime(&t));
	printf("2D %d point order %d rule "
		"(%d:%d:%d, %d tries, %d evals), error = %0.4lg\n", npoints, 
		degree, s3, s21, s111, tries / nprocs + 1,
		eval_count, (double)error);
    }

    i = j = 0;
    if (s3 > 0) {
	if (prtflag)
	    printf("    Perm3(%0.16lg)\n", (double)(1./3.L));
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Dup3(%0.16lg)\n", (double)W[j]);
	j++;
    }
    for (k = 0; k < s21; k++) {
	if (D[i] < -tol || D[i] > .5 + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Perm21(%0.16lg)\n", (double)D[i]);
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Dup21(%0.16lg)\n", (double)W[j]);
	i++;
	j++;
    }
    for (k = 0; k < s111; k++) {
	if (D[i] < -tol || D[i+1] < -tol || D[i] + D[i+1] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Perm111(%0.16lg, %0.16lg)\n",
			(double)D[i], (double)D[i + 1]);
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Dup111(%0.16lg)\n", (double)W[j]);
	i += 2;
	j += 1;
    }

    if (prtflag) {
	if (inbound) {
	    printf("Range check ok.\n");
	    fprintf(stderr, " :-:) got a good candidate (proc %d, %d tries, "
		    "%d evals, err = %0.3lg)\n", myrank, tries / nprocs + 1,
			eval_count, (double)error);
	}
	else {
	    printf(" *** abscissas out of range or weights negative!\n");
	    fprintf(stderr, " :-:( got a bad solution (%d:%d:%d, proc %d, "
		    "%d tries, %d evals, err = %0.3lg)\n", s3, s21, s111,
			myrank, tries / nprocs + 1,
			eval_count, (double)error);
	}
	fflush(stdout);
	fflush(stderr);
    }

    return inbound;
}

static void
search_2D(void)
{
    FLOAT error, *D, *W;
    int i, tries, printed = 0;

#if SYMMETRY
    nabsc = s21 + 2 * s111;	/* number of independent unknowns */
    nwgts = s3 + s21 + s111;	/* number of independent weights */
#else	/* SYMMETRY */
    nabsc = 2 * npoints;	/* number of independent unknowns */
    nwgts = npoints;		/* number of independent weights */
#endif	/* SYMMETRY */

    if (myrank == 0) {
	if (s3 > 0) fprintf(stderr, "%d*S3 ", s3);
	if (s21 > 0) fprintf(stderr, "%d*S21 ", s21);
	if (s111 > 0) fprintf(stderr, "%d*S111 ", s111);
	fprintf(stderr, "=> %d point(s), %d eqns, free absc: %d, "
			"free wgts: %d\n", npoints, neqns, nabsc, nwgts);
    }

    D = malloc(nabsc * sizeof(*D));
    W = malloc(nwgts * sizeof(*W));

#if ASYNCHRONOUS
    sync_file(REMOVE_SYNC_FILE);
    if (setjmp(jmp)) {
	check_found(0);
	goto end;
    }
#endif

    for (tries = 0; tries < (GENETIC ? 1:guess_max) + nprocs; tries += nprocs) {
#if !ASYNCHRONOUS
	if ((tries / nprocs + 1) % 5 == 0) {
	    if (check_found(0))
		goto end;
	}
#endif	/* !ASYNCHRONOUS */
	eval_count = 0;
#if GENETIC
	genetic(D);
#else
	abscissas_seed(D);
#endif
	if (plot) {
#if SYMMETRY
	    if (s21 > 0)
		plot_gen(s21, 0, (FLOAT)(1./3.L), D);
	    else
		plot_gen(2 * s111, s21, (FLOAT)1., D);
#else	/* SYMMETRY */
	    plot_gen(0, (FLOAT)1., D);
#endif	/* SYMMETRY */
	    break;;
	}

#if SYMMETRY
	error = nonlinear_lss(D, W);
	if (myrank == 0 && (tries / nprocs) % 1 == 0) {
	    fprintf(stderr, "    tries = %d, error = %lg, evals = %d\n",
			    tries, (double)error, eval_count);
	    fflush(stderr);
	}
	if (error < REPORT_TOL) {
	    i = check_last_found(D);
	    if (i == 1) {
		/* the solution has been found in last loop */
		continue;
	    }
	    else if (i >= 9) {
		/* the solution has been found in last two or more loops */
		fprintf(stderr, " *** got the same bad solution for the last "
				"10 times, switch to next case.\n");
#if ASYNCHRONOUS
		sync_file(CREATE_SYNC_FILE);
#endif	/* ASYNCHRONOUS */
		check_found(1);
		goto end;
	    }
	    if (check_solution_2d(0, tries, error, (FLOAT)0.5, D, W)) {
		if (check_solution_2d(1, tries, error, (FLOAT)1e-6, D, W)
			&& error < ACCEPT_TOL) {
		    printf("Seems to be a good solution, stop.\n");
		    fflush(stdout);
		    nsols++;
#if ASYNCHRONOUS
		    sync_file(CREATE_SYNC_FILE);
#endif	/* ASYNCHRONOUS */
		    check_found(0);
		    goto end;
		}
	    }
	    else if (!printed) {
		printed = 1;
		fprintf(stderr, " :-:( got a bad solution (%d:%d:%d, "
			"proc %d, %d tries, %d evals, err = %0.3lg)\n", 
			s3, s21, s111, myrank,
			tries / nprocs + 1, eval_count, (double)error);
	    }
	}
    }
#endif	/* SYMMETRY */

#if !ASYNCHROMOUS
    check_found(0);
#endif	/* !ASYNCHRONOUS */

end:
    free(W);
    free(D);
}

static int
check_solution_3d(int prtflag, int tries, FLOAT error, FLOAT tol,
			FLOAT D[], FLOAT W[])
{
    int inbound = 1, i, j, k;

#if !SYMMETRY
# error unimplemented!
#endif

    if (prtflag) {
	time_t t = time(NULL);
	printf("** %s", ctime(&t));
	printf("3D %d point order %d rule "
		"(%d:%d:%d:%d:%d, %d tries, %d evals), error = %0.4lg\n", 
		npoints, degree, s4, s31, s22, s211, s1111, tries / nprocs + 1,
		eval_count, (double)error);
    }

    i = j = 0;
    if (s4 > 0) {
	if (prtflag)
	    printf("    Perm4(%0.16lg)\n", 0.25);
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
 	}
	if (prtflag)
	    printf("    Dup4(%0.16lg)\n", (double)W[j]);
	j++;
    }

    for (k = 0; k < s31; k++) {
	if (D[i] < -tol || D[i] > 1./3.L + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Perm31(%0.16lg)\n", (double)D[i]);
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
	        printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Dup31(%0.16lg)\n", (double)W[j]);
	i++;
	j++;
    }

    for (k = 0; k < s22; k++) {
	if (D[i] < -tol || D[i] > .5 + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Perm22(%0.16lg)\n", (double)D[i]);
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
	        printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Dup22(%0.16lg)\n", (double)W[j]);
	i++;
	j++;
    }

    for (k = 0; k < s211; k++) {
	if (D[i] < -tol || D[i+1] < -tol || 2. * D[i] + D[i+1] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Perm211(%0.16lg, %0.16lg)\n",
			(double)D[i], (double)D[i + 1]);
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Dup211(%0.16lg)\n", (double)W[j]);
	i += 2;
	j += 1;
    }

    for (k = 0; k < s1111; k++) {
	if (D[i] < -tol || D[i+1] < -tol || D[i+2] < -tol ||
	    D[i] + D[i+1] + D[i+2] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Perm1111(%0.16lg, %0.16lg, %0.16lg)\n",
		(double)D[i], (double)D[i + 1], (double)D[i + 2]);
	if (W[j] < -tol || W[j] > 1. + tol) {
	    if (prtflag)
		printf("*");
	    inbound = 0;
	}
	if (prtflag)
	    printf("    Dup1111(%0.16lg)\n", (double)W[j]);
	i += 3;
	j += 1;
    }

    if (prtflag) {
	if (inbound) {
	    printf("Range check ok.\n");
	    fprintf(stderr, " :-:) got a good candidate (proc %d, %d tries, "
		    "%d evals, err = %0.3lg)\n", myrank, tries / nprocs + 1,
			eval_count, (double)error);
	}
	else {
	    printf(" *** abscissas out of range or weights negative!\n");
	    fprintf(stderr, " :-:( got a bad solution (%d:%d:%d:%d:%d, "
		    "proc %d, %d tries, %d evals, err = %0.3lg)\n", 
			s4, s31, s22, s211, s1111, myrank, tries / nprocs + 1,
			eval_count, (double)error);
	}
    }

    return inbound;
}

static void
search_3D(void)
/* evaluates 2D rule for the given symmetry pattern */
{
    FLOAT error, *D, *W;
    int i, tries, printed = 0;

#if SYMMETRY
    nabsc = s31 + s22 + 2 * s211 + 3 * s1111;
    nwgts = s4 + s31 + s22 + s211 + s1111;
#else	/* SYMMETRY */
    nabsc = 3 * npoints;	/* number of independent unknowns */
    nwgts = npoints;		/* number of independent weights */
#endif	/* SYMMETRY */

    if (myrank == 0) {
	if (s4 > 0) fprintf(stderr, "%d*S4 ", s4);
	if (s31 > 0) fprintf(stderr, "%d*S31 ", s31);
	if (s22 > 0) fprintf(stderr, "%d*S22 ", s22);
	if (s211 > 0) fprintf(stderr, "%d*S211 ", s211);
	if (s1111 > 0) fprintf(stderr, "%d*S1111 ", s1111);
	fprintf(stderr, "=> %d point(s), %d eqns, free absc: %d, "
			"free wgts: %d\n", npoints, neqns, nabsc, nwgts);
    }

    D = malloc(nabsc * sizeof(*D));
    W = malloc(nwgts * sizeof(*W));

#if ASYNCHRONOUS
    sync_file(REMOVE_SYNC_FILE);
    if (setjmp(jmp)) {
	check_found(0);
	goto end;
    }
#endif

    for (tries = 0; tries < (GENETIC ? 1:guess_max) + nprocs; tries += nprocs) {
#if !ASYNCHRONOUS
	if ((tries / nprocs + 1) % 5 == 0) {
	    if (check_found(0))
		goto end;
	}
#endif	/* !ASYNCHRONOUS */
	eval_count = 0;
#if GENETIC
	genetic(D);
#else
	abscissas_seed(D);
#endif
	if (plot) {
#if SYMMETRY
	    if (s31 > 0)
		plot_gen(s31, 0, (FLOAT)0.25, D);
	    else if (s22 > 0)
		plot_gen(s22, s31, (FLOAT)0.5, D);
	    else if (s211 > 0)
		plot_gen(2 * s211, s31 + s22, (FLOAT)1.0, D);
	    else
		plot_gen(3 * s1111, s31 + s22 + 2 * s211, (FLOAT)1.0, D);
#else	/* SYMMETRY */
	    plot_gen(0, (FLOAT)1.0, D);
#endif	/* SYMMETRY */
	    break;
	}

#if SYMMETRY
	error = nonlinear_lss(D, W);
	if (myrank == 0 && (tries / nprocs) % 1 == 0) {
	    fprintf(stderr, "    tries = %d, error = %lg, evals = %d\n",
			    tries, (double)error, eval_count);
	    fflush(stderr);
	}
	if (error < REPORT_TOL) {
	    i = check_last_found(D);
	    if (i == 1) {
		/* the solution has been found in last loop */
		continue;
	    }
	    else if (i >= 9) {
		/* the solution has been found in last two or more loops */
		fprintf(stderr, " *** got the same bad solution for the last "
				"10 times, switch to next case.\n");
#if ASYNCHRONOUS
		sync_file(CREATE_SYNC_FILE);
#endif	/* ASYNCHRONOUS */
		check_found(1);
		goto end;
	    }
	    if (check_solution_3d(0, tries, error, (FLOAT)0.5, D, W)) {
		if (check_solution_3d(1, tries, error, (FLOAT)1e-6, D, W)
			&& error < ACCEPT_TOL) {
		    printf("Seems to be a good solution, stop.\n");
		    fflush(stdout);
		    nsols++;
#if ASYNCHRONOUS
		    sync_file(CREATE_SYNC_FILE);
#endif	/* ASYNCHRONOUS */
		    check_found(0);
		    goto end;
		}
	    }
	    else if (!printed) {
		printed = 1;
		fprintf(stderr, " :-:( got a bad solution (%d:%d:%d:%d:%d, "
			"proc %d, %d tries, %d evals, err = %0.3lg)\n",
			s4, s31, s22, s211, s1111, myrank,
			tries / nprocs + 1, eval_count, (double)error);
	    }
	}
    }
#endif	/* SYMMETRY */

#if !ASYNCHROMOUS
    check_found(0);
#endif	/* !ASYNCHRONOUS */

end:
    free(W);
    free(D);
}

static void
search(int flag)
/* 'flag' is only used in process 0 (to tell slaves to stop if 'flag == 0') */
{
    if (specified_stars) {
	if (dim == 2)
	    search_2D();
	else
	    search_3D();
	return;
    }

    do {
#if USE_MPI
	int n, data[6];
	n = 0;
	if (myrank == 0) {
	    data[n++] = flag;
	    if (dim == 2) {
		data[n++] = s3;
		data[n++] = s21;
		data[n++] = s111;
	    }
	    else {
		data[n++] = s4;
		data[n++] = s31;
		data[n++] = s22;
		data[n++] = s211;
		data[n++] = s1111;
	    }
	}
	MPI_Bcast(data, dim == 2 ? 4 : 6, MPI_INT, 0, MPI_COMM_WORLD);
	if (myrank > 0) {
	    flag = data[n++];
	    if (dim == 2) {
		s3 = data[n++];
		s21 = data[n++];
		s111 = data[n++];
	    }
	    else {
		s4 = data[n++];
		s31 = data[n++];
		s22 = data[n++];
		s211 = data[n++];
		s1111 = data[n++];
	    }
	}
#endif

	if (!flag)
	    break;

	if (dim == 2)
	    search_2D();
	else
	    search_3D();
    } while (myrank > 0);
}

static void
permutations_2d(void)
/* decompose npoints into S3 + 3 S21 + 6 S111 */
{
    int i, j, n = npoints;

    s3 = n % 3;
    if (s3 == 2)
	return;
     n = (n - s3) / 3;
     j = n % 2;
     n /= 2;
     for (i = n; i >= 0; i--) {
	s21 = j + 2 * i;
	s111 = n - i;
	search(1);
	if (nsols > 0)
	    return;
    }
}

static void
permutations_3d(void)
/* decompose npoints into S4 + 4 S31 + 6 S22 + 12 S211 + 24 S1111 */
{
    int i, j, n = npoints;

    s4 = n % 2;
    n /= 2;
    /* decompose n into (2 S31 + 3 S22) + 6 (S211 + 2 S1111) = i + 6 * j */
    for (j = n / 6; j >= 0; j--) {
	i = n - 6 * j;
	/* decompose i into 2 * S31 + 3 * S22 */
	for (s22 = i / 3; s22 >= 0; s22--) {
	    s31 = i - 3 * s22;
	    if (s31 % 2 != 0)
		continue;
	    s31 /= 2;
	    /* and decompose j into S211 + 2 S1111 */
	    for (s1111 = j / 2; s1111 >= 0; s1111--) {
		s211 = j - 2 * s1111;
		search(1);
		if (nsols > 0)
		    return;
	    }
	}
    }
}

static void
permutations(void)
{
    if (dim == 2)
	permutations_2d();
    else
	permutations_3d();

    search(0);		/* tells slaves to leave */
}

int
main(int argc, char *argv[])
{
    int m;

#if 0
    FLOAT den0[den_n + 1];
    density_2D();
    memcpy(den0, den, sizeof(den0));
    free(den); den = NULL;
    density_3D();
    for (m = 0; m <= den_n; m++)
	printf("%lg %lg %lg %lg\n", (double)(m / (3.L * den_n)),
		(double)den0[m], (double)(m / (4.L * den_n)), (double)den[m]);
    exit(0);
#endif

#if USE_MPI
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
#endif

    if (argc > 3) {
	if (!strcmp(argv[argc - 1], "plot")) {
	    plot = 1;
	    argc--;
	}
	else if (!strcmp(argv[argc - 1], "print")) {
	    print = 1;
	    argc--;
	}
    }

    if (argc < 3) {
usage:
	fprintf(stderr, "Searching for a 2D rule:\n");
	fprintf(stderr, "    quadrule 2 order [[+]npoints [s3 s21 s111] "
			"[ntries]]"/*" ['plot'|'print']"*/"\n");
	fprintf(stderr, "Searching for a 3D rule:\n");
	fprintf(stderr, "    quadrule 3 order [[+]npoints "
			"[s4 s31 s22 s211 s1111] "
			"[ntries]]"/*" ['plot'|'print']"*/"\n");
	fprintf(stderr, "Example:\n    quadrule 3 5 14\n");
#if USE_MPI
	    MPI_Finalize();
#endif
	exit(1);
    }

    dim = atoi(argv[1]);
    degree = atoi(argv[2]);
    if (argc >= 4) {
	npoints = argv[3][0] == '+' ? -atoi(argv[3] + 1) : atoi(argv[3]);
    }
    if ((dim != 2 && dim != 3) || degree <= 0) {
usage1:
	fprintf(stderr, "\nError: invalid arguments\n");
	goto usage;
    }

#if HAVE_FEENABLEEXCEPT
    /* trap FP exceptions */
    feenableexcept(FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW);
#endif  /* HAVE_FEENABLEEXCEPT */

    if (argc > 4) {
	if (argc == 5) {
	    guess_max = atoi(argv[4]);
	}
	else if (dim == 2) {
	    if (argc != 7 && argc != 8)
		goto usage1;
	    specified_stars = 1;
	    s3 = atoi(argv[4]);
	    s21 = atoi(argv[5]);
	    s111 = atoi(argv[6]);
	    if (argc == 8)
		guess_max = atoi(argv[7]);
	    if (s3 != 0 && s3 != 1) {
		fprintf(stderr, "\nError: s3 can only be 0 or 1.\n");
		goto usage;
	    }
	    if (npoints != s3 + 3 * s21 + 6 * s111) {
		fprintf(stderr, "\nError: npoints and s3/s21/s111 mismatch.\n");
		goto usage;
	    }
	}
	else {
	    if (argc != 9 && argc != 10)
		goto usage1;
	    specified_stars = 1;
	    s4 = atoi(argv[4]);
	    s31 = atoi(argv[5]);
	    s22 = atoi(argv[6]);
	    s211 = atoi(argv[7]);
	    s1111 = atoi(argv[8]);
	    if (argc == 10)
		guess_max = atoi(argv[9]);
	    if (s4 != 0 && s4 != 1) {
		fprintf(stderr, "\nError: s4 can only be 0 or 1.\n");
		goto usage;
	    }
	    if (npoints != s4 + 4 * s31 + 6 * s22 + 12 * s211 + 24 * s1111) {
		fprintf(stderr, "\nError: npoints and s4/s31/s22/s211/s1111 "
				"mismatch.\n");
		goto usage;
	    }
	}
    }

    /* TODO: use a parallel random number generator */
    m = time(NULL);
#if USE_MPI
    MPI_Bcast(&m, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
    srand(m + myrank);

    /* compute the dimension of the space of polynomials of degree <= degree */
    if (dim == 2)
	npolys0 = (degree + 1) * (degree + 2) / (1 * 2);
    else
	npolys0 = (degree + 1) * (degree + 2) * (degree + 3) / (1 * 2 * 3);

    /* estimate minimum number of points */
    if (npoints <= 0) {
	npoints = (npolys0 + dim) / (dim + 1) - npoints;
#if 0
	if (dim == 2) {
	    if (degree & 1)
		npoints = (degree+1)*(degree+3)/8 + (degree+1)/4 - npoints;
	    else
		npoints = (degree+2)*(degree+4)/8 - npoints;
	}
	else {
	    /* TODO */
	}
#endif
    }

    /* get list of all polynomials */
    poly_gen();

    if (myrank == 0) {
	fprintf(stderr, "Dim=%d, deg=%d, npoints=%d, npolys=%d (%d)\n",
			dim, degree, npoints, npolys, npolys0);
#if INDEPENDENT_W
	fprintf(stderr, "Using lmder1 (weights as independent variables)\n");
#else
	fprintf(stderr, "Using lmdif1 (weights as dependent variables)\n");
#endif
    }

#if SYMMETRY
    /* determine the number of equations, which ensures exactness of the
     * quadrature rules for all polynomials <= some degree */
#if 0
    for (i = degree; (i + 1) * (i + 2) / 2 > n; i--) ;
    neqns = pdims[++i];
#else
    neqns = npolys;
#endif
#else	/* SYMMETRY */
    neqns = npoints;
#endif	/* SYMMETRY */

    assert(neqns <= npolys);

    if (myrank == 0 && !specified_stars)
	permutations();
    else {
	search(1);
    }

    if (pbasis != NULL) {
	for (m = 0; m < npolys; m++)
	    poly_free(pbasis[m]);
	free(pbasis);
    }
    if (porders != NULL)
	free(porders);
#if USE_CARDINAL && !ORTHOGONAL_BASIS
    if (obasis != NULL) {
	for (m = 0; m < npoints; m++)
	    poly_free(obasis[m]);
	free(obasis);
    }
    if (oorders != NULL)
	free(oorders);
#endif	/* USE_CARDINAL */
    if (pdims != NULL)
	free(pdims);

    if (myrank == 0)
	fprintf(stderr, "%d good solution(s) found.\n", nsols);

#if USE_MPI
    MPI_Finalize();
#endif

    return nsols > 0 ? 0 : 1;
}
